import copy
import logging
import random
from datetime import datetime, timedelta
from threading import Lock

from stem import ControllerError, DescriptorUnavailable, Flag
from stem.descriptor.router_status_entry import RouterStatusEntryV3
from stem.descriptor.server_descriptor import ServerDescriptor

from ..globals import (
    MAX_RECENT_CONSENSUS_COUNT,
    MAX_RECENT_PRIORITY_LIST_COUNT,
    MAX_RECENT_PRIORITY_RELAY_COUNT,
    MEASUREMENTS_PERIOD,
)
from ..util import timestamps

log = logging.getLogger(__name__)


def valid_after_from_network_statuses(network_statuses):
    """Obtain the consensus Valid-After datetime from the ``document``
    attribute of a ``stem.descriptor.RouterStatusEntryV3``.

    :param list network_statuses:

    returns datetime:
    """
    for ns in network_statuses:
        document = getattr(ns, "document", None)
        if document:
            valid_after = getattr(document, "valid_after", None)
            if valid_after:
                return valid_after
    return datetime.utcnow().replace(microsecond=0)


class Relay:
    def __init__(self, fp, cont, ns=None, desc=None, timestamp=None):
        """
        Given a relay fingerprint, fetch all the information about a relay that
        sbws currently needs and store it in this class. Acts as an abstraction
        to hide the confusion that is Tor consensus/descriptor stuff.

        :param str fp: fingerprint of the relay.
        :param cont: active and valid stem Tor controller connection

        :param datatime timestamp: the timestamp of a consensus
            (RouterStatusEntryV3) from which this relay has been obtained.
        """
        assert isinstance(fp, str)
        assert len(fp) == 40
        if ns is not None:
            assert isinstance(ns, RouterStatusEntryV3)
            self._ns = ns
        else:
            try:
                self._ns = cont.get_network_status(fp, default=None)
            except (DescriptorUnavailable, ControllerError) as e:
                log.exception("Exception trying to get ns %s", e)
                self._ns = None
        if desc is not None:
            assert isinstance(desc, ServerDescriptor)
            self._desc = desc
        else:
            try:
                self._desc = cont.get_server_descriptor(fp, default=None)
            except (DescriptorUnavailable, ControllerError) as e:
                log.exception("Exception trying to get desc %s", e)
        self.relay_in_recent_consensus = timestamps.DateTimeSeq(
            [], MAX_RECENT_CONSENSUS_COUNT
        )
        # Use the same timestamp as the consensus, so that it can be tested
        # that the relay was in a consensus using this timestamp.
        # Note that this doesn't change the number of consensus the relay was
        # in.
        self.update_relay_in_recent_consensus(timestamp)
        # The number of times that a relay is "prioritized" to be measured.
        # It is incremented in ``RelayPrioritizer.best_priority``
        self.relay_recent_priority_list = timestamps.DateTimeSeq(
            [], MAX_RECENT_PRIORITY_LIST_COUNT
        )
        # The number of times that a relay has been queued to be measured.
        # It is incremented in ``scanner.main_loop``
        self.relay_recent_measurement_attempt = timestamps.DateTimeSeq(
            [], MAX_RECENT_PRIORITY_LIST_COUNT
        )

    def _from_desc(self, attr):
        if not self._desc:
            return None
        return getattr(self._desc, attr, None)

    def _from_ns(self, attr):
        if not self._ns:
            return None
        return getattr(self._ns, attr, None)

    @property
    def nickname(self):
        return self._from_ns("nickname")

    @property
    def fingerprint(self):
        return self._from_ns("fingerprint")

    @property
    def flags(self):
        return self._from_ns("flags")

    @property
    def exit_policy(self):
        return self._from_desc("exit_policy")

    @property
    def average_bandwidth(self):
        return self._from_desc("average_bandwidth")

    @property
    def burst_bandwidth(self):
        return self._from_desc("burst_bandwidth")

    @property
    def observed_bandwidth(self):
        return self._from_desc("observed_bandwidth")

    @property
    def has_2_in_flowctrl(self):
        """
        Return True if the `FlowCtrl` field in the relay's protover consensus
        line include a value of 2 and False otherwise.

        """
        # NOTE: stem doesn't seem to obtain `pr`` nor create `protocols``
        # protocols = self._from_ns("protocols")
        protocols = self._from_desc("protocols")
        if protocols:
            if 2 in protocols.get("FlowCtrl", []):
                # log.debug("Exit %s has 2 in `FlowCtrl`.", self.nickname)
                return True
        # log.debug("Exit %s does not have 2 in `FlowCtrl`.", self.nickname)
        return False

    @property
    def consensus_bandwidth(self):
        """Return the consensus bandwidth in Bytes.

        Consensus bandwidth is the only bandwidth value that is in kilobytes.
        """
        if self._from_ns("bandwidth") is not None:
            return self._from_ns("bandwidth") * 1000

    @property
    def consensus_bandwidth_is_unmeasured(self):
        # measured appears only votes, unmeasured appears in consensus
        # therefore is_unmeasured is needed to know whether the bandwidth
        # value in consensus is coming from bwauth measurements or not.
        return self._from_ns("is_unmeasured")

    @property
    def address(self):
        return self._from_ns("address")

    @property
    def master_key_ed25519(self):
        """Obtain ed25519 master key of the relay in server descriptors.

        :returns: str, the ed25519 master key base 64 encoded without
                  trailing '='s.

        """
        # Even if this key is called master-key-ed25519 in dir-spec.txt,
        # it seems that stem parses it as ed25519_master_key
        key = self._from_desc("ed25519_master_key")
        if key is None:
            return None
        return key.rstrip("=")

    @property
    def consensus_valid_after(self):
        """Obtain the consensus Valid-After from the document of this relay
        network status.
        """
        network_status_document = self._from_ns("document")
        if network_status_document:
            return getattr(network_status_document, "valid_after", None)
        return None

    @property
    def last_consensus_timestamp(self):
        return self.relay_in_recent_consensus.last()

    def update_relay_in_recent_consensus(self, timestamp=None):
        self.relay_in_recent_consensus.update(timestamp)

    @property
    def relay_in_recent_consensus_count(self):
        """Number of times the relay was in a conensus."""
        return len(self.relay_in_recent_consensus)

    def can_exit_to_port(self, port, strict=False):
        """
        Returns True if the relay has an exit policy and the policy accepts
        exiting to the given port or False otherwise.

        If ``strict`` is true, it only returns the exits that can exit to all
        IPs and that port.

        The exits that are IPv6 only or IPv4 but rejecting some public networks
        will return false.
        On July 2020, there were 67 out of 1095 exits like this.

        If ``strict`` is false, it returns any exit that can exit to some
        public IPs and that port.

        Note that the EXIT flag exists when the relay can exit to 443 **and**
        80. Currently all Web servers are using 443, so it would not be needed
        to check the EXIT flag too, using this function.

        """
        assert isinstance(port, int)
        # if dind't get the descriptor, there isn't exit policy
        # When the attribute is gotten in getattr(self._desc, "exit_policy"),
        # is possible that stem's _input_rules is None and raises an exception
        # (#29899):
        #   File "/usr/lib/python3/dist-packages/sbws/lib/relaylist.py", line 117, in can_exit_to_port  # noqa
        #     if not self.exit_policy:
        #   File "/usr/lib/python3/dist-packages/stem/exit_policy.py", line 512, in __len__  # noqa
        #     return len(self._get_rules())
        #   File "/usr/lib/python3/dist-packages/stem/exit_policy.py", line 464, in _get_rules  # noqa
        #     for rule in decompressed_rules:
        # TypeError: 'NoneType' object is not iterable
        # Therefore, catch the exception here.
        try:
            if self.exit_policy:
                # Using `strip_private` to ignore reject rules to private
                # networks.
                # When ``strict`` is true, We could increase the chances that
                # the exit can exit via IPv6 too (``exit_policy_v6``). However,
                # in theory that is only known using microdescriptors.
                return self.exit_policy.strip_private().can_exit_to(
                    port=port, strict=strict
                )
        except TypeError:
            return False
        return False

    def is_exit_not_bad_allowing_port(self, port, strict=False):
        return (
            Flag.BADEXIT not in self.flags
            and Flag.EXIT in self.flags
            and self.can_exit_to_port(port, strict)
        )

    def increment_relay_recent_measurement_attempt(self):
        """
        Increment The number of times that a relay has been queued
        to be measured.

        It is call from :func:`~sbws.core.scaner.main_loop`.
        """
        self.relay_recent_measurement_attempt.update()

    @property
    def relay_recent_measurement_attempt_count(self):
        return len(self.relay_recent_measurement_attempt)

    def increment_relay_recent_priority_list(self):
        """
        The number of times that a relay is "prioritized" to be measured.

        It is call from
        :meth:`~sbws.lib.relayprioritizer.RelayPrioritizer.best_priority`.
        """
        # If it was not in the previous measurements version, start counting
        self.relay_recent_priority_list.update()

    @property
    def relay_recent_priority_list_count(self):
        return len(self.relay_recent_priority_list)

    # XXX: tech-debt: replace `_desc` attr by a a `dequee` of the last
    # descriptors seen for this relay and the timestamp.
    def update_server_descriptor(self, server_descriptor):
        """Update this relay server descriptor (from the consensus."""
        self._desc = server_descriptor

    # XXX: tech-debt: replace `_ns` attr by a a `dequee` of the last
    # router statuses seen for this relay and the timestampt.
    def update_router_status(self, router_status):
        """Update this relay router status (from the consensus)."""
        self._ns = router_status


class RelayList:
    """Keeps a list of all relays in the current Tor network and updates it
    transparently in the background. Provides useful interfaces for getting
    only relays of a certain type.
    """

    def __init__(
        self,
        args,
        conf,
        controller,
        measurements_period=MEASUREMENTS_PERIOD,
        state=None,
    ):
        self._controller = controller
        self.rng = random.SystemRandom()
        self._refresh_lock = Lock()
        # To track all the consensus seen.
        self._recent_consensus = timestamps.DateTimeSeq(
            [], MAX_RECENT_CONSENSUS_COUNT, state, "recent_consensus"
        )
        # Initialize so that there's no error trying to access to it.
        # In future refactor, change to a dictionary, where the keys are
        # the relays' fingerprint.
        self._relays = []
        # The period of time for which the measurements are keep.
        self._measurements_period = measurements_period
        self._recent_measurement_attempt = timestamps.DateTimeSeq(
            [],
            MAX_RECENT_PRIORITY_RELAY_COUNT,
            state,
            "recent_measurement_attempt",
        )
        # Start with 0 for the min bw for our second hops
        self._exit_min_bw = 0
        self._non_exit_min_bw = 0
        self._refresh()

    def _need_refresh(self):
        # New consensuses happen every hour.
        return datetime.utcnow() >= self.last_consensus_timestamp + timedelta(
            seconds=60 * 60
        )

    @property
    def last_consensus_timestamp(self):
        """Returns the datetime when the last consensus was obtained."""
        return self._recent_consensus.last()

    @property
    def relays(self):
        # See if we can get the list of relays without having to do a refresh,
        # which is expensive and blocks other threads
        if self._need_refresh():
            log.debug(
                "We need to refresh our list of relays. "
                "Going to wait for lock."
            )
            # Whelp we couldn't just get the list of relays because the list is
            # stale. Wait for the lock so we can refresh it.
            with self._refresh_lock:
                log.debug(
                    "We got the lock. Now to see if we still "
                    "need to refresh."
                )
                # Now we have the lock ... but wait! Maybe someone else already
                # did the refreshing. So check if it still needs refreshing. If
                # not, we can do nothing.
                if self._need_refresh():
                    log.debug("Yup we need to refresh our relays. Doing so.")
                    self._refresh()
                else:
                    log.debug(
                        "No we don't need to refresh our relays. "
                        "It was done by someone else."
                    )
            log.debug("Giving back the lock for refreshing relays.")
        return self._relays

    @property
    def fast(self):
        return self._relays_with_flag(Flag.FAST)

    @property
    def exits(self):
        return self._relays_with_flag(Flag.EXIT)

    @property
    def bad_exits(self):
        return self._relays_with_flag(Flag.BADEXIT)

    @property
    def non_exits(self):
        return self._relays_without_flag(Flag.EXIT)

    @property
    def guards(self):
        return self._relays_with_flag(Flag.GUARD)

    @property
    def authorities(self):
        return self._relays_with_flag(Flag.AUTHORITY)

    @property
    def relays_fingerprints(self):
        # Using relays instead of _relays, so that the list get updated if
        # needed, since this method is used to know which fingerprints are in
        # the consensus.
        return [r.fingerprint for r in self.relays]

    def random_relay(self):
        return self.rng.choice(self.relays)

    def _relays_with_flag(self, flag):
        return [r for r in self.relays if flag in r.flags]

    def _relays_without_flag(self, flag):
        return [r for r in self.relays if flag not in r.flags]

    def _init_relays(self):
        """Returns a new list of relays that are in the current consensus.
        And update the consensus timestamp list with the current one.

        """
        c = self._controller
        # This will get router statuses from this Tor cache, might not be
        # updated with the network.
        # Change to stem.descriptor.remote in future refactor.
        network_statuses = c.get_network_statuses()
        new_relays_dict = dict([(r.fingerprint, r) for r in network_statuses])
        log.debug(
            "Number of relays in the current consensus: %d.",
            len(new_relays_dict),
        )

        # Find the timestamp of the last consensus.
        timestamp = valid_after_from_network_statuses(network_statuses)
        self._recent_consensus.update(timestamp)

        new_relays = []

        # Only or debugging, count the relays that are not in the current
        # consensus and have not been seen in the last consensuses either.
        num_old_relays = 0

        relays = copy.deepcopy(self._relays)
        for r in relays:
            if r.fingerprint in new_relays_dict.keys():
                # If a relay in the previous consensus and is in the current
                # one, update its timestamp, router status and descriptor.
                fp = r.fingerprint
                # new_relays_dict[fp] is the router status.
                r.update_router_status(new_relays_dict[fp])
                r.update_relay_in_recent_consensus(timestamp)
                try:
                    descriptor = c.get_server_descriptor(fp, default=None)
                except (DescriptorUnavailable, ControllerError) as e:
                    log.exception("Exception trying to get desc %s", e)
                r.update_server_descriptor(descriptor)
                # Add it to the new list of relays.
                new_relays.append(r)
                # And remove it from the new consensus dict, as it has
                # already added to the new list.
                new_relays_dict.pop(fp)

            # In #30727, the relay that is not in the current conensus but is
            # not "old", was added to the new list of relays too.
            # In #40037 we think it should not be measured, as it might cause
            # many circuit errors. It's already added to the generator.
            # Otherwise, don't add it to the new list of relays.
            # For debugging, count the old relays that will be discarded.
            else:
                num_old_relays += 1

        # Finally, add the relays that were not in the previous consensus
        for fp, ns in new_relays_dict.items():
            r = Relay(ns.fingerprint, c, ns=ns, timestamp=timestamp)
            new_relays.append(r)

        days = self._measurements_period / (60 * 60 * 24)
        log.debug(
            "Previous number of relays being measured %d", len(self._relays)
        )
        log.debug(
            "Number of relays not in the in the consensus in the last "
            "%d days: %d.",
            days,
            num_old_relays,
        )
        log.debug(
            "Number of relays to measure with the current consensus: " "%d",
            len(new_relays),
        )
        return new_relays

    def _refresh(self):
        # Set a new list of relays.
        self._relays = self._init_relays()

        log.info(
            "Number of consensuses obtained in the last %s days: %s.",
            int(self._measurements_period / 24 / 60 / 60),
            self.recent_consensus_count,
        )

        # Calculate minimum bandwidth value for 2nd hop after we refreshed
        # our available relays.
        self._calculate_min_bw_second_hop()
        self.set_consensus_params()

    @property
    def recent_consensus_count(self):
        """Number of times a new consensus was obtained."""
        return len(self._recent_consensus)

    def exits_with_2_in_flowctrl(self, port):
        """
        Return the exits that include a value of 2 in the `FlowCtrl` field
        in their protover consensus line.

        """
        exits_flowctrl2 = [
            r
            for r in self.exits_not_bad_allowing_port(port)
            if r.has_2_in_flowctrl
        ]
        # In chutney, just take relays with exit flag
        if (
            not exits_flowctrl2
            and self._controller.get_conf("TestingTorNetwork") == "1"
        ):
            exits_flowctrl2 = list(
                filter(lambda r: r.has_2_in_flowctrl, self.exits)
            )
        return exits_flowctrl2

    def exits_without_2_in_flowctrl(self, port):
        """
        Return the exits that do not include a value of 2 in the `FlowCtrl`
        field in their protover consensus line or the field is missing.

        """
        return [
            r
            for r in self.exits_not_bad_allowing_port(port)
            if not r.has_2_in_flowctrl
        ]

    def set_consensus_params(self):
        """Obtain current consensus params fields and store them as an attr.

        It is not possible to obtain them from `get_network_statuses` via
        control port, only via a cached file.

        """
        if self._controller.get_conf("TestingTorNetwork") == "1":
            log.debug("In a testing network.")
            self.consensus_params_dict = {}
            return
        log.debug("Not in a testing network.")
        consensus = self._controller.get_info(
            "dir/status-vote/current/consensus"
        )
        from unittest import mock

        if isinstance(consensus, mock.Mock):
            log.debug("Mocked consensus.")
            self.consensus_params_dict = {}
            return
        # Create a dictionary from all the consensus lines.
        consensus_dict = dict(
            [
                (line.split(" ")[0], line.split()[1:])
                for line in consensus.split("\n")
            ]
        )
        # Create a dictionary from the consensus `params` line.
        self.consensus_params_dict = dict(
            [
                (p.split("=")[0], int(p.split("=")[1]))
                for p in consensus_dict.get("params", [])
            ]
        )
        log.debug("Consensus params: %s", self.consensus_params_dict.items())

    @property
    def is_consensus_cc_alg_2(self):
        """
        Return True if the consensus document has a value of 2 in the `cc_alg`
        field.

        From proposals/324-rtt-congestion-control.txt spec::

            6.5.1. Parameters common to all algorithms
            [...]
            cc_alg:
            - Description:
                Specifies which congestion control algorithm clients should
                use, as an integer.
            - Range: [0,3]  (0=fixed, 1=Westwood, 2=Vegas, 3=NOLA)
            - Default: 2

        """
        if (
            self.consensus_params_dict
            and self.consensus_params_dict.get("cc_alg", 0) == 2
        ):
            log.info("The consensus implements congestion control.")
            return True
        log.info("The consensus does not implement congestion control.")
        return False

    @property
    def is_consensus_bwscanner_cc_gte_1(self):
        """
        Return True if the consensus document has a value of 1 or greater in
        the `bwscanner_cc` field."""
        if (
            self.consensus_params_dict
            and self.consensus_params_dict.get("bwscanner_cc", 0) >= 1
        ):
            log.info(
                "The consensus says to use exits that support congestion"
                " control."
            )
            return True
        log.info(
            "The consensus says to use exits that do not support congestion"
            " control."
        )
        return False

    def exits_not_bad_allowing_port(self, port, strict=False):
        return [
            r
            for r in self.exits
            if r.is_exit_not_bad_allowing_port(port, strict)
        ]

    def increment_recent_measurement_attempt(self):
        """
        Increment the number of times that any relay has been queued to be
        measured.

        It is call from :func:`~sbws.core.scaner.main_loop`.

        It is read and stored in a ``state`` file.
        """
        # NOTE: blocking, writes to file!
        self._recent_measurement_attempt.update()

    @property
    def recent_measurement_attempt_count(self):
        return len(self._recent_measurement_attempt)

    def _calculate_min_bw_second_hop(self):
        """
        Calculates the minimum bandwidth for both exit and non-exit relays
        chosen as a second hop by picking the lowest bandwidth value available
        from the top 75% of the respective category.
        """
        # Sort our sets of candidates according to bw, lowest amount first.
        # It's okay to keep things simple for the calculation and go over all
        # exits, including badexits.
        exit_candidates = sorted(
            self.exits, key=lambda r: r.consensus_bandwidth
        )
        non_exit_candidates = sorted(
            self.non_exits, key=lambda r: r.consensus_bandwidth
        )
        # We know the bandwidth is sorted from least to most. Dividing the
        # length of the available relays by 4 gives us the position of the
        # relay with the lowest bandwidth from the top 75%. We do this both
        # for our exit and non-exit candidates.
        pos = int(len(exit_candidates) / 4)
        self._exit_min_bw = exit_candidates[pos].consensus_bandwidth
        pos = int(len(non_exit_candidates) / 4)
        # when there are not non-exits in a test network
        if pos:
            self._non_exit_min_bw = non_exit_candidates[
                pos
            ].consensus_bandwidth

    def exit_min_bw(self):
        return self._exit_min_bw

    def non_exit_min_bw(self):
        return self._non_exit_min_bw

    @property
    def sum_consensus_bw(self):
        return sum(
            list(
                map(
                    lambda r: r.consensus_bandwidth,
                    self.relays,
                )
            )
        )

    @property
    def sum_consensus_bw_exits_not_bad_allowing_port(self):
        return sum(
            list(
                map(
                    lambda r: r.consensus_bandwidth,
                    self.exits_not_bad_allowing_port(443),
                )
            )
        )

    @property
    def sum_consensus_bw_exits_flowctrl2(self):
        return sum(
            list(
                map(
                    lambda r: r.consensus_bandwidth,
                    self.exits_with_2_in_flowctrl(443),
                )
            )
        )



import logging

from stem import (
    CircuitExtensionFailed,
    ControllerError,
    InvalidArguments,
    InvalidRequest,
    ProtocolError,
    SocketClosed,
    Timeout,
)

log = logging.getLogger(__name__)


def valid_circuit_length(path):
    return 0 < len(path) <= 8


class CircuitBuilder:
    """The CircuitBuilder interface.

    Subclasses must implement their own build_circuit() function.
    Subclasses may keep additional state if they'd find it helpful.

    The primary way to use a CircuitBuilder of any type is to simply create it
    and then call cb.build_circuit(...) with any options that your
    CircuitBuilder type needs.

    It might be good practice to close circuits as you find you no longer need
    them, but CircuitBuilder will keep track of existing circuits and close
    them when it is deleted.
    """

    # XXX: In new major version, remove args and conf, they are not used.
    def __init__(
        self,
        args,
        conf,
        controller,
        relay_list=None,
        close_circuits_on_exit=True,
    ):
        self.controller = controller
        self.built_circuits = set()
        self.close_circuits_on_exit = close_circuits_on_exit
        self.circuit_timeout = conf.getint("general", "circuit_timeout")

    def close_circuit(self, circ_id):
        try:
            self.controller.close_circuit(circ_id)
        # SocketClosed will be raised when stopping sbws
        except (InvalidArguments, InvalidRequest, SocketClosed) as e:
            log.debug(e)
        self.built_circuits.discard(circ_id)

    def _build_circuit_impl(self, path):
        """
        :returns tuple: circuit id if the circuit was built, error if there
            was an error building the circuit.
        """
        if not valid_circuit_length(path):
            return None, "Can not build a circuit, invalid path."
        c = self.controller
        timeout = self.circuit_timeout
        fp_path = "[" + " -> ".join([p for p in path]) + "]"
        log.debug("Building %s", fp_path)
        try:
            circ_id = c.new_circuit(path, await_build=True, timeout=timeout)
        except (
            InvalidRequest,
            CircuitExtensionFailed,
            ProtocolError,
            Timeout,
            SocketClosed,
        ) as e:
            return None, str(e)
        return circ_id, None

    def __del__(self):
        c = self.controller
        if not self.close_circuits_on_exit:
            return
        for circ_id in self.built_circuits:
            try:
                c.get_circuit(circ_id, default=None)
                try:
                    c.close_circuit(circ_id)
                except (InvalidArguments, InvalidRequest):
                    pass
            except (ControllerError, InvalidArguments) as e:
                log.exception(
                    "Exception trying to get circuit to delete: %s", e
                )
        self.built_circuits.clear()


# In a future refactor, remove this class, since sbws chooses the relays to
# build the circuit, the relays are not just chosen as random as this class
# does.
class GapsCircuitBuilder(CircuitBuilder):
    """Same as ``CircuitBuilder`` but implements build_circuit."""

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    def build_circuit(self, path):
        """Return parent class build circuit method.

        Since sbws is only building 2 hop paths, there is no need to add random
        relays to the path, or convert back and forth between fingerprint and
        ``Relay`` objects.

        """
        return self._build_circuit_impl(path)



"""Util classes to manipulate sequences of datetime timestamps.

Optionally update also a state file.

"""
# Workarounds to store datetimes for objects because they are not compossed
# by other objects nor stored in a database with a creation datetime.
import collections
import logging
from datetime import datetime, timedelta

from sbws.util.timestamp import is_old

log = logging.getLogger(__name__)


class DateTimeSeq(collections.deque):
    """Store and manage a datetime sequence and optionally a state file."""

    def __init__(self, iterable=[], maxlen=None, state=None, state_key=None):
        self._maxlen = maxlen
        self._items = collections.deque(iterable, maxlen)
        self._state = state
        self._state_key = state_key

    def _remove_old(self):
        self._items = collections.deque(
            filter(lambda x: not is_old(x), self._items), maxlen=self._maxlen
        )

    def update(self, dt=None):
        self._remove_old()
        self._items.append(dt or datetime.utcnow().replace(microsecond=0))
        if self._state is not None and self._state_key:
            self._state[self._state_key] = list(self._items)
        return list(self._items)

    def last(self):
        if len(self._items) > 0:
            return self._items[-1]
        return datetime.utcnow().replace(microsecond=0) - timedelta(hours=1)

    def list(self):
        return list(self._items)

    def __len__(self):
        return len(self._items)


class DateTimeIntSeq(collections.deque):
    """
    Store and manage a sequence of lists composed of a datetime and an int.

    Optionally store and manage an state file.
    """

    def __init__(self, iterable=[], maxlen=None, state=None, state_key=None):
        self._maxlen = maxlen
        self._items = collections.deque(iterable, maxlen)
        self._state = state
        self._state_key = state_key

    def _remove_old(self):
        self._items = collections.deque(
            filter(lambda x: not is_old(x[0]), self._items),
            maxlen=self._maxlen,
        )

    def update(self, dt=None, number=0):
        self._remove_old()
        # Because json serializes tuples to lists, use list instead of tuple
        # to facilitate comparisons.
        self._items.append(
            [dt or datetime.utcnow().replace(microsecond=0), number]
        )
        if self._state is not None and self._state_key:
            self._state[self._state_key] = list(self._items)
        return list(self._items)

    def last(self):
        if len(self._items) > 0:
            return self._items[-1]
        return datetime.utcnow().replace(microsecond=0) - timedelta(hours=1)

    def list(self):
        return list(self._items)

    def __len__(self):
        return sum(map(lambda x: x[1], self._items))



"""Integration tests for destination.py"""
import sbws.util.requests as requests_utils
from sbws.lib.destination import (
    Destination,
    DestinationList,
    connect_to_destination_over_circuit,
)


def test_destination_list_no_usability_test_success(
    conf, persistent_launch_tor, cb, rl
):
    # In a future refactor, if DestionationList is not initialized with the
    # controller, this test should be an unit test.
    destination_list, error_msg = DestinationList.from_config(
        conf, cb, rl, persistent_launch_tor
    )
    # Because there's only 1 destination in conftest, random should return
    # the same one.
    assert destination_list.next() == destination_list._all_dests[0]


def test_connect_to_destination_over_circuit_success(
    persistent_launch_tor, dests, cb, rl
):
    destination = dests.next()
    session = requests_utils.make_session(persistent_launch_tor, 10)
    # Choose a relay that is not an exit
    relay = [r for r in rl.relays if r.nickname == "test005m"][0]
    # Choose an exit, for this test it does not matter the bandwidth
    helper = rl.exits_not_bad_allowing_port(destination.port)[0]
    circuit_path = [relay.fingerprint, helper.fingerprint]
    # build a circuit
    circuit_id, _ = cb.build_circuit(circuit_path)
    # Perform "usability test"
    is_usable, response = connect_to_destination_over_circuit(
        destination, circuit_id, session, persistent_launch_tor, 1024
    )
    assert is_usable is True
    assert "content_length" in response
    assert destination.is_functional()


def test_connect_to_destination_over_circuit_fail(
    persistent_launch_tor, dests, cb, rl
):
    bad_destination = Destination("https://example.example", 1024, False)
    session = requests_utils.make_session(persistent_launch_tor, 10)
    # Choose a relay that is not an exit
    relay = [r for r in rl.relays if r.nickname == "test005m"][0]
    # Choose an exit, for this test it does not matter the bandwidth
    helper = rl.exits_not_bad_allowing_port(bad_destination.port)[0]
    circuit_path = [relay.fingerprint, helper.fingerprint]
    # Build a circuit.
    circuit_id, _ = cb.build_circuit(circuit_path)
    # Perform "usability test"
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )
    assert is_usable is False

    # because it is the first time it fails, failures aren't count
    assert bad_destination.is_functional()

    # fail three times in a row
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )
    assert not bad_destination.is_functional()


def test_functional_destinations(conf, cb, rl, persistent_launch_tor):
    good_destination = Destination("https://127.0.0.1:28888", 1024, False)
    bad_destination = Destination("https://example.example", 1024, False)

    session = requests_utils.make_session(persistent_launch_tor, 10)
    # Choose a relay that is not an exit
    relay = [r for r in rl.relays if r.nickname == "test005m"][0]
    # Choose an exit, for this test it does not matter the bandwidth
    helper = rl.exits_not_bad_allowing_port(bad_destination.port)[0]
    circuit_path = [relay.fingerprint, helper.fingerprint]
    # Build a circuit.
    circuit_id, _ = cb.build_circuit(circuit_path)

    # fail three times in a row
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )
    is_usable, response = connect_to_destination_over_circuit(
        bad_destination, circuit_id, session, persistent_launch_tor, 1024
    )

    destination_list = DestinationList(
        conf,
        [good_destination, bad_destination],
        cb,
        rl,
        persistent_launch_tor,
    )
    functional_destinations = destination_list.functional_destinations
    assert [good_destination] == functional_destinations



import logging
from argparse import ArgumentDefaultsHelpFormatter

import sbws.util.stem as stem_utils
from sbws.lib.relaylist import RelayList
from sbws.util.state import State

log = logging.getLogger(__name__)


def gen_parser(sub):
    d = "Log the number of exits that have 2 in FlowCtrl."
    p = sub.add_parser(
        "flowctrl2",
        description=d,
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    return p


def main(args, conf):
    controller = stem_utils.launch_or_connect_to_tor(conf)

    # When there will be a refactor where conf is global, this can be removed
    # from here.
    state = State(conf.getpath("paths", "state_fname"))

    measurements_period = conf.getint("general", "data_period")
    rl = RelayList(args, conf, controller, measurements_period, state)
    exits = rl.exits_not_bad_allowing_port(443)
    log.info("Number of exits: %s", len(exits))
    exits_min_bw = rl.exit_min_bw()
    log.info("Exits minimum bandwidth: %s KB.", exits_min_bw / 1000)
    exits_with_min_bw = stem_utils.only_relays_with_bandwidth(
        controller, exits, min_bw=exits_min_bw
    )
    log.info(
        "Number of exits with minimum bandwidth: %s", len(exits_with_min_bw)
    )
    exits_sorted = sorted(
        exits, key=lambda r: r.consensus_bandwidth, reverse=True
    )
    log.info(
        "Exits lowest bandwidth: %s KB.",
        exits_sorted[-1].consensus_bandwidth / 1000,
    )
    log.info(
        "Exits highest bandwidth: %s KB.",
        exits_sorted[0].consensus_bandwidth / 1000,
    )

    non_exits = rl.non_exits
    log.info("Number of non exits: %s.", len(non_exits))

    non_exits_with_helpers_double_bw = 0
    non_exits_with_helpers_same_bw = 0
    non_exits_without_helpers_same_double_bw = 0
    for relay in non_exits:
        double_min_bw = max(exits_min_bw, relay.consensus_bandwidth * 2)
        helpers = stem_utils.only_relays_with_bandwidth(
            exits_with_min_bw, min_bw=double_min_bw
        )
        if helpers:
            log.debug(
                "Number of helpers with double bandwidth for relay %s: %s.",
                relay.nickname,
                len(helpers),
            )
            non_exits_with_helpers_double_bw += 1
        else:
            min_bw = max(exits_min_bw, relay.consensus_bandwidth)
            helpers = stem_utils.only_relays_with_bandwidth(
                exits_with_min_bw, min_bw=min_bw
            )
            if helpers:
                log.debug(
                    "Number of helpers for relay %s: %s.",
                    relay.nickname,
                    len(helpers),
                )
                non_exits_with_helpers_same_bw += 1
            else:
                log.debug("No helpers for relay %s", relay.nickname)
                non_exits_without_helpers_same_double_bw += 1
    log.info(
        "Number of non exits with helpers that have double bandwidth: %s.",
        non_exits_with_helpers_double_bw,
    )
    log.info(
        "Number of non exits with helpers that have same bandwidth: %s.",
        non_exits_with_helpers_same_bw,
    )
    log.info(
        "Number of non exits without helpers that have  double or same"
        " bandwidth: %s.",
        non_exits_without_helpers_same_double_bw,
    )

    exits_flowctrl2 = rl.exits_with_2_in_flowctrl(443)
    log.info(
        "Number of exits that have 2 in FlowCtrl: %s.", len(exits_flowctrl2)
    )
    exits_flowctrl2_min_bw = stem_utils.only_relays_with_bandwidth(
        controller, exits_flowctrl2, min_bw=exits_min_bw
    )
    log.info(
        "Number of exits that have 2 in FlowCtrl and minimum bandwidth: %s",
        len(exits_flowctrl2_min_bw),
    )
    exits_flowctrl2_sorted = sorted(
        exits_flowctrl2, key=lambda r: r.consensus_bandwidth, reverse=True
    )
    log.info(
        "Exits that have 2 in FlowCtrl lowest bandwidth: %s KB.",
        exits_flowctrl2_sorted[-1].consensus_bandwidth / 1000,
    )
    log.info(
        "Exits that have 2 in FlowCtrl highest bandwidth: %s KB.",
        exits_flowctrl2_sorted[0].consensus_bandwidth / 1000,
    )

    non_exits_with_helpers_flowctrl2_double_bw = 0
    non_exits_with_helpers_flowctrl2_same_bw = 0
    non_exits_without_helpers_flowctrl2_same_double_bw = 0
    for relay in non_exits:
        double_min_bw = max(exits_min_bw, relay.consensus_bandwidth * 2)
        helpers = stem_utils.only_relays_with_bandwidth(
            controller, exits_flowctrl2_min_bw, min_bw=double_min_bw
        )
        if helpers:
            log.debug(
                "Number of helpers with double bandwidth for relay %s: %s.",
                relay.nickname,
                len(helpers),
            )
            non_exits_with_helpers_flowctrl2_double_bw += 1
        else:
            min_bw = max(exits_min_bw, relay.consensus_bandwidth)
            helpers = stem_utils.only_relays_with_bandwidth(
                controller, exits_flowctrl2_min_bw, min_bw=min_bw
            )
            if helpers:
                log.debug(
                    "Number of helpers for relay %s: %s.",
                    relay.nickname,
                    len(helpers),
                )
                non_exits_with_helpers_flowctrl2_same_bw += 1
            else:
                log.debug("No helpers for relay %s", relay.nickname)
                non_exits_without_helpers_flowctrl2_same_double_bw += 1
    log.info(
        "Number of non exits with helpers that have 2 in FlowCtrl and double"
        " bandwidth: %s.",
        non_exits_with_helpers_flowctrl2_double_bw,
    )
    log.info(
        "Number of non exits with helpers that have 2 in FlowCtrl and same"
        " bandwidth: %s.",
        non_exits_with_helpers_flowctrl2_same_bw,
    )
    log.info(
        "Number of non exits without helpers that have 2 in FlowCtrl and"
        " double or same bandwidth: %s.",
        non_exits_without_helpers_flowctrl2_same_double_bw,
    )

    sum_consensus_bw = rl.sum_consensus_bw
    log.info("Total consensus weight: %s", sum_consensus_bw / 1000)
    sum_consensus_bw_exits_not_bad_allowing_port = (
        rl.sum_consensus_bw_exits_not_bad_allowing_port
    )
    log.info(
        "Consensus weight of exits (without BAD flag, allowing 443 port): %s",
        sum_consensus_bw_exits_not_bad_allowing_port / 1000,
    )
    sum_consensus_bw_exits_flowctrl2 = rl.sum_consensus_bw_exits_flowctrl2
    log.info(
        "Cnsensus weight exits (without BAD flag, allowing 443 port)"
        " with 2 in FlowCtrl: %s",
        sum_consensus_bw_exits_flowctrl2,
    )
    fraction_flowctrl2_exits = (
        sum_consensus_bw_exits_flowctrl2
        / sum_consensus_bw_exits_not_bad_allowing_port
    )
    log.info(
        "Fraction of consensus weight of exits with 2 in FlowCtrl with respect"
        " all the exits: %.2f",
        fraction_flowctrl2_exits,
    )



"""Util functions to convert between timestamp formats"""
from datetime import datetime, timedelta

from ..globals import MEASUREMENTS_PERIOD


def dt_obj_to_isodt_str(dt):
    """
    Convert datetime object to ISO 8601 string.

    :param datetime dt: datetime object in UTC timezone
    :returns: ISO 8601 string
    """
    assert isinstance(dt, datetime)
    # Using naive datetime object without timezone, assumed utc
    return dt.replace(microsecond=0).isoformat()


def isostr_to_dt_obj(isostr):
    return datetime.strptime(isostr, "%Y-%m-%dT%H:%M:%S")


def unixts_to_dt_obj(unixts):
    """
    Convert unix timestamp to naive datetime object in UTC time zone.

    :param float/int/str unixts: unix timestamp
    :returns: datetime object in UTC timezone
    """
    if isinstance(unixts, str):
        try:
            unixts = int(unixts)
        except ValueError:
            unixts = float(unixts)
    if isinstance(unixts, float):
        unixts = int(unixts)
    assert isinstance(unixts, int)
    return datetime.utcfromtimestamp(unixts)


def unixts_to_isodt_str(unixts):
    """
    Convert unix timestamp to ISO 8601 string in UTC time zone.

    :param float/int/str unixts: unix timestamp
    :returns: ISO 8601 string in UTC time zone
    """
    return dt_obj_to_isodt_str(unixts_to_dt_obj(unixts))


def now_unixts():
    return datetime.utcnow().timestamp()


def now_isodt_str():
    """Return datetime now as ISO 8601 string in UTC time zone."""
    return dt_obj_to_isodt_str(datetime.utcnow())


def now_fname():
    """
    Return now timestamp in UTC formatted as %Y%m%d_%H%M%S string for file
    names.

    :returns: now timestamp in UTC formatted as %Y%m%d_%H%M%S string
    """
    return datetime.utcnow().strftime("%Y%m%d_%H%M%S")


def unixts_to_str(unixts):
    """Convert unix timestamp integer or float to string"""
    # even if it is only converting to str, ensure that input is nothing else
    # than int or float
    assert isinstance(unixts, int) or isinstance(unixts, float)
    return str(unixts)


# XXX: tech-debt: replace all the code that check whether a
# measurement or relay is older than the measurement period by this.
def is_old(timestamp, measurements_period=MEASUREMENTS_PERIOD):
    """Whether the given timestamp is older that the given measurements
    period.
    """
    if not isinstance(timestamp, datetime):
        if isinstance(timestamp, str):
            # This will raise an exception if the string is not correctly
            # formatted.
            timestamp = isostr_to_dt_obj(timestamp)
        elif isinstance(timestamp, int) or isinstance(timestamp, float):
            # This will raise an exception if the type is not int or float or
            # is not actually a timestamp
            timestamp = unixts_to_dt_obj(timestamp)
    oldest_date = datetime.utcnow() - timedelta(seconds=measurements_period)
    return timestamp < oldest_date



import collections
import datetime
import logging
import random
from urllib.parse import urlparse

import requests
from stem.control import EventType

import sbws.util.stem as stem_utils
from sbws import settings
from sbws.globals import DESTINATION_VERIFY_CERTIFICATE

from ..globals import (
    DELTA_SECONDS_RETRY_DESTINATION,
    FACTOR_INCREMENT_DESTINATION_RETRY,
    MAX_NUM_DESTINATION_FAILURES,
    MAX_SECONDS_RETRY_DESTINATION,
    NUM_DESTINATION_ATTEMPTS_STORED,
)

log = logging.getLogger(__name__)


# Duplicate some code from DestinationList.from_config,
# it should be refactored.
def parse_destinations_countries(conf):
    """Returns the destinations' country as string separated by comma."""
    destinations_countries = []
    for key in conf["destinations"].keys():
        # Not a destination key
        if key in ["usability_test_interval"]:
            continue
        # The destination is not enabled
        if not conf["destinations"].getboolean(key):
            continue
        destination_section = "destinations.{}".format(key)
        destination_country = conf[destination_section].get("country", None)
        destinations_countries.append(destination_country)
    return ",".join(destinations_countries)


def _parse_verify_option(conf_section):
    if "verify" not in conf_section:
        return DESTINATION_VERIFY_CERTIFICATE
    try:
        verify = conf_section.getboolean("verify")
    except ValueError:
        log.warning(
            "Currently sbws only supports verify=true/false, not a CA bundle "
            "file. We think %s is not a bool, and thus must be a CA bundle "
            "file. This is supposed to be allowed by the Python Requests "
            "library, but pastly couldn't get it to work in his afternoon "
            "of testing. So we will allow this, but expect Requests to throw "
            "SSLError exceptions later. Have fun!",
            conf_section["verify"],
        )
        return conf_section["verify"]
    if not verify:
        # disable urllib3 warning: InsecureRequestWarning
        import urllib3

        urllib3.disable_warnings()
    return verify


def connect_to_destination_over_circuit(dest, circ_id, session, cont, max_dl):
    """
    Connect to **dest* over the given **circ_id** using the given Requests
    **session**. Make sure the destination seems usable. Return True and a
    dictionary of helpful information if we connected and the destination is
    usable.  Otherwise return False and a string stating what the issue is.

    This function has two effects, and which one is the "side effect" depends
    on your goal.

    1. It creates a stream to the destination. It persists in the requests
    library **session** object so future requests use the same stream.
    Therefore, the primary effect of this function could be to open a
    connection to the destination that measurements can be made over the given
    **circ_id**, which makes the usability checks a side effect (yet important
    sanity check).

    2. It determines if a destination is usable. Therefore, the primary effect
    of this function could be to perform the usability checks and return the
    results of those checks, which makes the persistent stream a side effect
    that we don't care about.

    As of the time of writing, you'll find that sbws/core/scanner.py uses this
    function in order to obtain that stream over which to perform measurements.
    You will also find in sbws/lib/destination.py (this file) this function
    being used to determine if a Destination is usable. The first relies on the
    persistent stream side effect, the second ignores it (and in fact throws it
    away when it closes the circuit).

    :param dest Destination: the place to which we should connect
    :param circ_id str: the circuit we should connect over
    :param session Session: the Requests library session object to use to make
        the connection.
    :param cont Controller: them Stem library controller controlling Tor
    :returns: True and a dictionary if everything is in order and measurements
        should commence.  False and an error string otherwise.
    """
    assert isinstance(dest, Destination)
    log.debug("Connecting to destination over circuit.")
    # Do not start if sbws is stopping
    if settings.end_event.is_set():
        return False, "Shutting down."
    error_prefix = "When sending HTTP HEAD to {}, ".format(dest.url)
    with stem_utils.stream_building_lock:
        listener = stem_utils.attach_stream_to_circuit_listener(cont, circ_id)
        stem_utils.add_event_listener(cont, listener, EventType.STREAM)
        try:
            head = session.head(dest.url, verify=dest.verify)
        except requests.exceptions.RequestException as e:
            dest.add_failure()
            return False, "Could not connect to {} over circ {} {}: {}".format(
                dest.url, circ_id, stem_utils.circuit_str(cont, circ_id), e
            )
        finally:
            stem_utils.remove_event_listener(cont, listener)
    if head.status_code != requests.codes.ok:
        dest.add_failure()
        return (
            False,
            error_prefix + "we expected HTTP code "
            "{} not {}".format(requests.codes.ok, head.status_code),
        )
    if "content-length" not in head.headers:
        dest.add_failure()
        return (
            False,
            error_prefix + "we expect the header Content-Length "
            "to exist in the response",
        )
    content_length = int(head.headers["content-length"])
    if max_dl > content_length:
        dest.add_failure()
        return (
            False,
            error_prefix + "our maximum configured download size "
            "is {} but the content is only {}".format(max_dl, content_length),
        )
    log.debug("Connected to %s over circuit %s", dest.url, circ_id)
    # Any failure connecting to the destination will call add_failure,
    # It can not be set at the start, to be able to know whether it is
    # failing consecutive times.
    dest.add_success()
    return True, {"content_length": content_length}


class Destination:
    """Web server from which data is downloaded to measure bandwidth."""

    # NOTE: max_dl and verify should be optional and have defaults
    def __init__(
        self,
        url,
        max_dl,
        verify,
        max_num_failures=MAX_NUM_DESTINATION_FAILURES,
        delta_seconds_retry=DELTA_SECONDS_RETRY_DESTINATION,
        max_seconds_between_retries=MAX_SECONDS_RETRY_DESTINATION,
        num_attempts_stored=NUM_DESTINATION_ATTEMPTS_STORED,
        factor_increment_retry=FACTOR_INCREMENT_DESTINATION_RETRY,
    ):
        """Initializes the Web server from which the data is downloaded.

        :param str url: Web server data URL to download.
        :param int max_dl: Maximum size of the the data to download.
        :param bool verify: Whether to verify or not the TLS certificate.
        :param int max_num_failures: Number of consecutive failures when the
            destination is not considered functional.
        :param int delta_seconds_retry: Delta time to try a destination
            that was not functional.
        :param int num_attempts_stored: Number of attempts to store.
        :param int factor_increment_retry: Factor to increment delta by
            before trying to use a destination again.
        """
        self._max_dl = max_dl
        u = urlparse(url)
        self._url = u
        self._verify = verify

        # Attributes to decide whether a destination is functional or not.
        self._max_num_failures = max_num_failures
        self._num_attempts_stored = num_attempts_stored
        # Default delta time to try a destination that was not functional.
        self._default_delta_seconds_retry = delta_seconds_retry
        self._delta_seconds_retry = delta_seconds_retry
        # A cap on the time to wait between destination retries.
        self._max_seconds_between_retries = max_seconds_between_retries
        # Using a deque (FIFO) to do not grow forever and
        # to do not have to remove old attempts.
        # Store tuples of timestamp and whether the destination succeed or not
        # (succeed, 1, failed, 0).
        # Initialize it as if it never failed.
        self._attempts = collections.deque(
            [
                (datetime.datetime.utcnow(), 1),
            ],
            maxlen=self._num_attempts_stored,
        )
        self._factor = factor_increment_retry

    def _last_attempts(self, n=None):
        """Return the last ``n`` attempts the destination was used."""
        # deque does not accept slices,
        # a new deque is returned with the last n items
        # (or less if there were less).
        return collections.deque(
            self._attempts, maxlen=(n or self._max_num_failures)
        )

    def _are_last_attempts_failures(self, n=None):
        """
        Return True if the last`` n`` times the destination was used
        and failed.
        """
        # Count the number that there was a failure when used
        n = n if n else self._max_num_failures
        return [i[1] for i in self._last_attempts(n)].count(
            0
        ) >= self._max_num_failures

    def _increment_time_to_retry(self, factor=None):
        """
        Increment the time a destination will be tried again by a ``factor``.
        """
        self._delta_seconds_retry *= factor or self._factor
        if self._delta_seconds_retry > self._max_seconds_between_retries:
            self._delta_seconds_retry = self._max_seconds_between_retries
            log.debug(
                "Incremented the time to try destination %s past the "
                "limit, capping it at %s hours.",
                self.url,
                self._delta_seconds_retry / 60 / 60,
            )
        else:
            log.debug(
                "Incremented the time to try destination %s to %s " "hours.",
                self.url,
                self._delta_seconds_retry / 60 / 60,
            )

    def _get_last_try_in_seconds_ago(self):
        """
        Return the delta between the last try and now, as positive seconds.
        """
        # Timestamp of the last attempt.
        last_time = self._attempts[-1][0]
        return (datetime.datetime.utcnow() - last_time).total_seconds()

    def _is_last_try_old_enough(self, n=None):
        """
        Return True if the last time it was used it was ``n`` seconds ago.
        """
        # If the last attempt is older than _delta_seconds_retry, try again
        return self._get_last_try_in_seconds_ago() > self._delta_seconds_retry

    def is_functional(self):
        """Whether connections to a destination are failing or not.

        Return True if:
            - It did not fail more than n (by default 3) consecutive times.
            - The last time the destination was tried
              was x (by default 3h) seconds ago.
        And False otherwise.

        When the destination is tried again after the consecutive failures,
        the time to try again is incremented and reset as soon as the
        destination does not fail.
        """
        # NOTE: does a destination fail because several threads are using
        # it at the same time?
        # If a destination fails for 1 minute and there're 3 threads, the
        # 3 threads will fail.

        # Failed the last X consecutive times
        if self._are_last_attempts_failures():
            # The log here will appear in all the the queued relays and
            # threads.
            log.debug(
                "The last %s times the destination %s failed. "
                "It last ran %s seconds ago. "
                "Disabled for %s seconds.",
                self._max_num_failures,
                self.url,
                self._get_last_try_in_seconds_ago(),
                self._delta_seconds_retry,
            )
            log.warning(
                "The last %s times a destination failed. "
                "It last ran %s seconds ago. "
                "Disabled for %s seconds."
                "Please, add more destinations or increment the "
                "number of maximum number of consecutive failures "
                "in the configuration.",
                self._max_num_failures,
                self._get_last_try_in_seconds_ago(),
                self._delta_seconds_retry,
            )
            # It was not used for a while and the last time it was used
            # was long ago, then try again
            if self._is_last_try_old_enough():
                log.debug(
                    "The destination %s was not tried for %s seconds, "
                    "it is going to by tried again.",
                    self.url,
                    self._get_last_try_in_seconds_ago(),
                )
                # Set the next time to retry higher, in case this attempt fails
                self._increment_time_to_retry()
                return True
            return False
        # Reset the time to retry to the initial value
        # In case it was incrememented
        self._delta_seconds_retry = self._default_delta_seconds_retry
        return True

    def add_failure(self, dt=None):
        self._attempts.append((dt or datetime.datetime.utcnow(), 0))

    def add_success(self, dt=None):
        self._attempts.append((dt or datetime.datetime.utcnow(), 1))

    @property
    def url(self):
        return self._url.geturl()

    @property
    def verify(self):
        return self._verify

    @property
    def hostname(self):
        return self._url.hostname

    @property
    def port(self):
        p = self._url.port
        scheme = self._url.scheme
        if p is None:
            if scheme == "http":
                p = 80
            elif scheme == "https":
                p = 443
            else:
                assert None, "Unreachable. Unknown scheme {}".format(scheme)
        assert p is not None
        return p

    @staticmethod
    def from_config(conf_section, max_dl, number_threads):
        assert "url" in conf_section
        url = conf_section["url"]
        verify = _parse_verify_option(conf_section)
        try:
            # Because one a destination fails, all the threads that are using
            # it at that moment will fail too, multiply by the number of
            # threads.
            max_num_failures = (
                conf_section.getint("max_num_failures")
                or MAX_NUM_DESTINATION_FAILURES
            )
        except ValueError:
            # If the operator did not setup the number, set to the default.
            max_num_failures = MAX_NUM_DESTINATION_FAILURES

        max_num_failures *= number_threads
        return Destination(url, max_dl, verify, max_num_failures)


class DestinationList:
    def __init__(self, conf, dests, circuit_builder, relay_list, controller):
        assert len(dests) > 0
        for dest in dests:
            assert isinstance(dest, Destination)
        self._rng = random.SystemRandom()
        self._cont = controller
        self._cb = circuit_builder
        self._rl = relay_list
        self._all_dests = dests

    @property
    def functional_destinations(self):
        return [d for d in self._all_dests if d.is_functional()]

    @staticmethod
    def from_config(conf, circuit_builder, relay_list, controller):
        assert "destinations" in conf
        section = conf["destinations"]
        dests = []
        for key in section.keys():
            if key in ["usability_test_interval"]:
                continue
            if not section.getboolean(key):
                log.debug("%s is disabled; not loading it", key)
                continue
            dest_sec = "destinations.{}".format(key)
            assert dest_sec in conf  # validate_config should require this
            log.debug("Loading info for destination %s", key)
            dests.append(
                Destination.from_config(
                    conf[dest_sec],
                    # Multiply by the number of threads since all the threads
                    # will fail at the same time.
                    conf.getint("scanner", "max_download_size"),
                    conf.getint("scanner", "measurement_threads"),
                )
            )
        if len(dests) < 1:
            msg = (
                "No enabled destinations in config. Please see "
                'docs/source/man_sbws.ini.rst" or "man 5 sbws.ini" '
                "for help adding and enabling destinations."
            )
            return None, msg
        return (
            DestinationList(
                conf, dests, circuit_builder, relay_list, controller
            ),
            "",
        )

    def next(self):
        """
        Returns the next destination that should be used in a measurement
        """
        # Do not perform usability tests since a destination is already proven
        # usable or not in every measurement, and it should depend on a X
        # number of failures.
        # This removes the need for an extra lock for every measurement.
        # Do not change the order of the destinations, just return a
        # destination.
        # random.choice raises IndexError with an empty list.
        if self.functional_destinations:
            return self._rng.choice(self.functional_destinations)
        else:
            return None



import logging

import pytest

from sbws.core.scanner import (
    _pick_ideal_second_hop,
    measure_relay,
    select_helper_candidates,
    use_relay_as_entry,
)
from sbws.lib.resultdump import ResultSuccess


def assert_within(value, target, radius):
    """
    Assert that **value** is within **radius** of **target**

    If target is 10 and radius is 2, value can be anywhere between 8 and 12
    inclusive
    """
    assert (
        target - radius < value
    ), "Value is too small. {} is not within " "{} of {}".format(
        value, radius, target
    )
    assert (
        target + radius > value
    ), "Value is too big. {} is not within " "{} of {}".format(
        value, radius, target
    )


@pytest.mark.skip(
    reason=(
        "Disabled because chutney is not creating a network"
        "with relay1mbyteMAB."
    )
)
def test_measure_relay_with_maxadvertisedbandwidth(
    persistent_launch_tor, sbwshome_dir, args, conf, dests, cb, rl, caplog
):
    caplog.set_level(logging.DEBUG)
    # d = get_everything_to_measure(sbwshome, cont, args, conf)
    # rl = d['rl']
    # dests = d['dests']
    # cb = d['cb']
    # 117A456C911114076BEB4E757AC48B16CC0CCC5F is relay1mbyteMAB
    relay = [r for r in rl.relays if r.nickname == "relay1mbyteMAB"][0]
    # d['relay'] = relay
    result = measure_relay(args, conf, dests, cb, rl, relay)
    assert len(result) == 1
    result = result[0]
    assert isinstance(result, ResultSuccess)
    one_mbyte = 1 * 1024 * 1024
    dls = result.downloads
    for dl in dls:
        # This relay has MaxAdvertisedBandwidth set, but should not be limited
        # to just 1 Mbyte. Assume and assert that all downloads where at least
        # more than 10% faster than 1 MBps
        assert dl["amount"] / dl["duration"] > one_mbyte * 1.1
    assert result.relay_average_bandwidth == one_mbyte


@pytest.mark.skip(reason="temporally disabled")
def test_measure_relay_with_relaybandwidthrate(
    persistent_launch_tor, args, conf, dests, cb, rl
):
    relay = [r for r in rl.relays if r.nickname == "relay1mbyteRBR"][0]
    result = measure_relay(args, conf, dests, cb, rl, relay)
    assert len(result) == 1
    result = result[0]
    assert isinstance(result, ResultSuccess)
    one_mbyte = 1 * 1024 * 1024
    allowed_error = 0.1 * one_mbyte  # allow 10% error in either direction
    dls = result.downloads
    for dl in dls:
        assert_within(dl["amount"] / dl["duration"], one_mbyte, allowed_error)


def test_second_hop_has_2_in_flowctrl(
    is_cc_tor_version, dests, rl, persistent_launch_tor
):
    if not is_cc_tor_version:
        import pytest

        pytest.skip("This test can't be run with this tor version")
        return
    rl.consensus_params_dict = {"cc_alg": 2, "bwscanner_cc": 1}
    assert rl.is_consensus_cc_alg_2
    assert rl.is_consensus_bwscanner_cc_gte_1
    dest = dests._all_dests[0]
    relay = rl._relays[0]

    relay_as_entry = use_relay_as_entry(relay, rl, dest)
    candidates = select_helper_candidates(relay, rl, dest, relay_as_entry)
    helper = _pick_ideal_second_hop(relay, rl, relay_as_entry, candidates)
    assert helper.has_2_in_flowctrl



from unittest.mock import patch

from sbws import settings
from sbws.lib.relayprioritizer import RelayPrioritizer
from sbws.lib.resultdump import ResultDump, ResultErrorCircuit, ResultSuccess


def static_time(value):
    while True:
        yield value


def _build_result_for_relay(conf, rl, result_type, relay_nick, timestamp):
    relay = [r for r in rl.relays if r.nickname == relay_nick]
    assert len(relay) == 1
    relay = relay[0]
    other = [r for r in rl.relays if r.nickname != relay_nick][0]
    circ = [relay.fingerprint, other.fingerprint]
    rtts = [0.5, 0.5, 0.5]
    dls = [
        {"amount": 1024, "duration": 1},
        {"amount": 1024, "duration": 1},
        {"amount": 1024, "duration": 1},
    ]
    if result_type == ResultSuccess:
        return ResultSuccess(
            rtts,
            dls,
            relay,
            circ,
            conf["destinations.foo"]["url"],
            "test",
            t=timestamp,
        )

    elif result_type == ResultErrorCircuit:
        return ResultErrorCircuit(
            relay,
            circ,
            conf["destinations.foo"]["url"],
            "test",
            msg="Test error circ message",
            t=timestamp,
        )


@patch("time.time")
def test_relayprioritizer_general(
    time_mock, sbwshome_empty, args, conf, rl, persistent_launch_tor
):
    now = 1000000
    time_mock.side_effect = static_time(now)
    rd = ResultDump(args, conf)
    try:
        rp = RelayPrioritizer(args, conf, rl, rd)
        results = [
            _build_result_for_relay(
                conf,
                rl,
                ResultSuccess,
                "test{:03d}m".format(i),
                now - (i * 100),
            )
            # In chutney the relays are from 003 to 011
            for i in range(3, 12)
        ]
        for result in results:
            rd.store_result(result)
        best_list = list(rp.best_priority())
        # With chutney, the relays not measured, with higher priority, will be
        # the 3 exits and authorities.
        # So take the list from the first measured relay, ie. from the 6th
        # position.
        # The measured relays will be in inverse order to their name.
        best_list_measured = best_list[6:]
        for i in range(3, 12):
            nick = "test{:03d}m".format(i)
            # -1 To start by the back, - 2 because their names start by 3,
            # not 1
            pos = (i - 2) * -1
            relay = best_list_measured[pos]
            assert relay.nickname == nick
            assert relay.relay_recent_priority_list_count == 1
    finally:
        settings.end_event.set()



#!/usr/bin/env python3
# File: scale-v3bw-with-budget.py
# Written by: Matt Traudt
# Copyright/License: CC0
from collections import OrderedDict
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys


def fail_hard(*s):
    print(*s, file=sys.stderr)
    exit(1)


def line_into_dict(line):
    words = line.strip().split()
    d = OrderedDict()
    for word in words:
        key, value = word.split('=')
        d.update({key: value})
    return d


def main(args):
    total_input_weight = 0
    line_dicts = []
    for line in args.input:
        if 'node_id=' not in line:
            args.output.write(line)
            continue
        d = line_into_dict(line)
        # Check that the required parts of the line are here
        if 'node_id' not in d:
            fail_hard('Line without required node_id:', line)
        if 'bw' not in d:
            fail_hard('Line without required bw:', line)
        # Make sure the bw looks like an int
        try:
            d['bw'] = int(d['bw'])
        except ValueError as e:
            fail_hard('Found a non-int bw value', d['bw'], e)
        # Accumulate the total "bandwidth" weights on the input side
        total_input_weight += d['bw']
        # And keep the line for later
        line_dicts.append(d)
    # Now calculate a ratio to multiply every line by. It's the total budget we
    # should give ourselves based on the number of relays in the v3bw file (AKA
    # the total output weight) divided by the total input weight
    ratio = (len(line_dicts) * args.budget_per_relay) / total_input_weight
    for d in line_dicts:
        d['bw'] = round(d['bw'] * ratio)
        # Accumulate all the parts of the line back together
        s = ''
        for key in d:
            s += '{}={} '.format(key, d[key])
        # Remove trailing ' ' and replace with '\n'
        s = s.rstrip() + '\n'
        args.output.write(s)


def gen_parser():
    d = 'Read a v3bw file, adjust the bandwidth weights of the relays, and '\
        'write the new v3bw file out. For each relay in the v3bw file, we '\
        'give ourselves some amount of weight to work with. We then '\
        'distribute this weight to the relays in the same proportions their '\
        'input weights were in. This cases the scale of theie weights to '\
        'move up or down, but their relative weights stay the same.'
    p = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                       description=d)
    p.add_argument('-i', '--input', type=FileType('rt'),
                   default='/dev/stdin',
                   help='Input v3bw file to be scaled')
    p.add_argument('-o', '--output', type=FileType('wt'),
                   default='/dev/stdout',
                   help='Where to write a new, and scaled, v3bw file')
    p.add_argument('--budget-per-relay', type=float, default=7500,
                   help='Per relay in the v3bw file, add this much to our '
                   'budget')
    return p


if __name__ == '__main__':
    p = gen_parser()
    args = p.parse_args()
    exit(main(args))



"""relayprioritizer.py unit tests."""
from freezegun import freeze_time


def test_increment_recent_priority_list(relay_prioritizer):
    """Test that incrementing the priority lists do not go on forever.

    And instead it only counts the number of priority lists in the last days.
    """

    state = relay_prioritizer._state
    assert 0 == relay_prioritizer.recent_priority_list_count
    assert not state.get("recent_priority_list", None)

    # Pretend that a priority list is made.
    with freeze_time("2020-02-29 10:00:00"):
        relay_prioritizer.increment_recent_priority_list()
    assert 1 == relay_prioritizer.recent_priority_list_count
    assert 1 == len(state["recent_priority_list"])

    # And a second priority list is made 4 days later.
    with freeze_time("2020-03-04 10:00:00"):
        relay_prioritizer.increment_recent_priority_list()
    assert 2 == relay_prioritizer.recent_priority_list_count
    assert 2 == len(state["recent_priority_list"])

    # And a third priority list is made 5 days later.
    with freeze_time("2020-03-05 10:00:00"):
        relay_prioritizer.increment_recent_priority_list()
    assert 3 == relay_prioritizer.recent_priority_list_count
    assert 3 == len(state["recent_priority_list"])

    # And a fourth priority list is made 6 days later. The first one is
    # now removed and not counted.
    with freeze_time("2020-03-06 10:00:00"):
        relay_prioritizer.increment_recent_priority_list()
    assert 3 == relay_prioritizer.recent_priority_list_count
    assert 3 == len(state["recent_priority_list"])


def test_increment_priority_relay(relay_prioritizer):
    """Test that incrementing the number of relays in the priority lists
    do not go on forever.

    And instead it only counts number of relays in priority lists in the last
    days.
    """

    state = relay_prioritizer._state
    assert 0 == relay_prioritizer.recent_priority_relay_count
    assert not state.get("recent_priority_relay", None)

    # Pretend that a priority list is made.
    with freeze_time("2020-02-29 10:00:00"):
        relay_prioritizer.increment_recent_priority_relay(2)
    assert 2 == relay_prioritizer.recent_priority_relay_count
    assert 2 == state.count("recent_priority_relay")

    # And a second priority list is made 4 days later.
    with freeze_time("2020-03-04 10:00:00"):
        relay_prioritizer.increment_recent_priority_relay(2)
    assert 4 == relay_prioritizer.recent_priority_relay_count
    assert 4 == state.count("recent_priority_relay")

    # And a third priority list is made 5 days later.
    with freeze_time("2020-03-05 10:00:00"):
        relay_prioritizer.increment_recent_priority_relay(2)
    assert 6 == relay_prioritizer.recent_priority_relay_count
    assert 6 == state.count("recent_priority_relay")

    # And a fourth priority list is made 6 days later. The first one is
    # now removed and the relays are not counted.
    with freeze_time("2020-03-06 10:00:00"):
        relay_prioritizer.increment_recent_priority_relay(2)
    assert 6 == relay_prioritizer.recent_priority_relay_count
    assert 6 == state.count("recent_priority_relay")



#!/usr/bin/env python3
# File: get-per-relay-budget.py
# Written by: Matt Traudt
# Copyright/License: CC0
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from statistics import median
from statistics import mean
from stem.control import Controller
import stem


def _get_controller_port(args):
    return Controller.from_port(port=args.ctrl_port)


def _get_controller_socket(args):
    return Controller.from_socket_file(path=args.ctrl_socket)


def get_controller(args):
    try:
        cont = _get_controller_port(args)
    except stem.SocketError:
        cont = _get_controller_socket(args)
    return cont


def print_quiet(bws):
    print(round(mean(bws)))


def print_regular(bws):
    print(len(bws), 'relays')
    print('mean:', round(mean(bws)))
    print('median:', round(median(bws)))


def main(args):
    cont = get_controller(args)
    cont.authenticate()
    bws = [ns.bandwidth for ns in cont.get_network_statuses()]
    if args.quiet:
        print_quiet(bws)
    else:
        print_regular(bws)


def gen_parser():
    d = 'Get the consensus weight for every relay in the current consensus '\
        'and print some information about them.'
    p = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                       description=d)
    p.add_argument('--ctrl-port', metavar='PORT', type=int, default=9051,
                   help='Port on which to control the Tor client')
    p.add_argument('--ctrl-socket', metavar='SOCK', type=str,
                   default='/var/run/tor/control',
                   help='Path to socket on which to control the Tor client')
    p.add_argument('-q', '--quiet', action='store_true',
                   help='If given, only print the mean bandwidth to stdout')
    return p


if __name__ == '__main__':
    p = gen_parser()
    args = p.parse_args()
    exit(main(args))



import fcntl
import logging
import os

from sbws.globals import fail_hard

log = logging.getLogger(__name__)


class _FLock:
    def __init__(self, lock_fname):
        self._lock_fname = lock_fname
        self._fd = None

    def __enter__(self):
        mode = os.O_RDWR | os.O_CREAT | os.O_TRUNC
        self._fd = os.open(self._lock_fname, mode)
        log.debug("Going to lock %s", self._lock_fname)
        try:
            fcntl.flock(self._fd, fcntl.LOCK_EX)
        except OSError as e:
            fail_hard(
                "We couldn't call flock. Are you on an unsupported "
                "platform? Error: %s",
                e,
            )
        log.debug("Received lock %s", self._lock_fname)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._fd is not None:
            log.debug("Releasing lock %s", self._lock_fname)
            os.close(self._fd)


class DirectoryLock(_FLock):
    """
    Holds a lock on a file in **dname** so that other sbws processes/threads
    won't try to read/write while we are reading/writing in this directory.

    >>> with DirectoryLock(dname):
    >>>     # do some reading/writing in dname
    >>> # no longer have the lock

    Note: The directory must already exist.

    :param str dname: Name of directory for which we want to obtain a lock
    """

    def __init__(self, dname):
        assert os.path.isdir(dname)
        lock_fname = os.path.join(dname, ".lockfile")
        super().__init__(lock_fname)


class FileLock(_FLock):
    """
    Holds a lock on **fname** so that other sbws processes/threads
    won't try to read/write while we are reading/writing this file.

    >>> with FileLock(fname):
    >>>     # do some reading/writing of fname
    >>> # no longer have the lock

    :param str fname: Name of the file for which we want to obtain a lock
    """

    def __init__(self, fname):
        lock_fname = fname + ".lockfile"
        super().__init__(lock_fname)



"""
Integration tests for the files with data to be used by the bandwidth file.

"""
from sbws.lib.resultdump import load_recent_results_in_datadir
from sbws.lib.v3bwfile import V3BWFile
from sbws.util.state import State


def test_results(conf):
    results = load_recent_results_in_datadir(5, conf["paths"]["datadir"])
    for fp, values in results.items():
        count = max(
            [
                len(getattr(r, "relay_recent_measurement_attempt", []))
                for r in values
            ]
        )
        assert count == 1
        count = max(
            [len(getattr(r, "relay_in_recent_consensus", [])) for r in values]
        )
        assert count == 1
        count = max(
            [len(getattr(r, "relay_recent_priority_list", [])) for r in values]
        )
        assert count == 1


def test_state(conf):
    state = State(conf["paths"]["state_fpath"])
    assert 1 == state.count("recent_consensus")
    assert 1 == state.count("recent_priority_list")
    assert 15 == state.count("recent_priority_relay")
    # Because of 40023, it's 0. Change to 15 if we store it again at some point
    # assert 15 == state.count("recent_measurement_attempt")
    assert not state.count("recent_measurement_attempt")


def test_v3bwfile(conf):
    bwfile = V3BWFile.from_v1_fpath(
        conf["paths"]["v3bw_fname"].format("latest")
    )
    assert "1" == bwfile.header.recent_consensus_count
    assert "1" == bwfile.header.recent_priority_list_count
    assert "15" == bwfile.header.recent_priority_relay_count
    # Because of 40023, there is not this header
    # assert "15" == bwfile.header.recent_measurement_attempt_count
    for bwline in bwfile.bw_lines:
        assert 1 == bwline.relay_in_recent_consensus_count
        assert 1 == bwline.relay_recent_priority_list_count
        assert 1 == bwline.relay_recent_measurement_attempt_count



import os

from sbws.util.state import State

# from tempfile import NamedTemporaryFile as NTF


def test_state_set_allowed_key_types(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    attempt_keys = "k"
    for key in attempt_keys:
        state[key] = 4
        assert state[key] == 4


def test_state_set_allowed_value_types(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    attempt_vals = (15983, None, True, -1.2, "loooooool")
    for val in attempt_vals:
        state["foo"] = val
        assert state["foo"] == val


def test_state_del(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    d = {"a": 1, "b": 2, "c": 3, "d": 4}
    for key in d:
        state[key] = d[key]
    assert len(state) == len(d)

    del d["a"]
    del state["a"]
    assert len(state) == len(d)
    for key in d:
        assert d[key] == state[key]

    d["e"] = 5
    state["e"] = 5
    d["e"] = 5.5
    state["e"] = 5.5
    assert len(state) == len(d)


def test_state_get_len(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    d = {"a": 1, "b": 2, "c": 3, "d": 4}
    for key in d:
        state[key] = d[key]
    assert len(state) == len(d)

    del d["a"]
    del state["a"]
    assert len(state) == len(d)

    d["e"] = 5
    state["e"] = 5
    d["e"] = 5.5
    state["e"] = 5.5
    assert len(state) == len(d)


def test_state_contains(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    d = {"a": 1, "b": 2, "c": 3, "d": 4}
    for key in d:
        state[key] = d[key]
    assert "a" in state
    assert "e" not in state


def test_state_iter(tmpdir):
    state = State(os.path.join(str(tmpdir), "statefoo"))
    for key in state:
        pass
    d = {"a": 1, "b": 2, "c": 3, "d": 4}
    for key in d:
        state[key] = d[key]
    assert set([key for key in state]) == set(d)


def test_two_instances(tmpdir):
    """Test that 2 different instances don't overwrite each other"""
    s1 = State(os.path.join(str(tmpdir), "state.dat"))
    s2 = State(os.path.join(str(tmpdir), "state.dat"))
    s1["x"] = "foo"
    s2["y"] = "bar"
    assert s2["x"] == "foo"


def test_datetime_values(tmpdir):
    import datetime

    state = State(os.path.join(str(tmpdir), "state.dat"))
    now = datetime.datetime.utcnow().replace(microsecond=0)
    state["datetimes"] = now
    assert now == state["datetimes"]



#!/usr/bin/env python3
# File: sbws-http-server.py
# Author: Matt Traudt
# License: CC0
#
# This script implements just enough of the HTTP protocol to work with Simple
# Bandwidth Scanner.
#
# All requested URLs exist. All return 1 GiB of garbage data. We always speak
# HTTP/1.1 because that's necessary for Keep-Alive request headers
# (used by sbws scanners) to work.
#
# HEAD and GET requests are supported to the minimum extent necessary.
# This essentially means that if the client sends Range request headers just
# like sbws does, then we'll only send back the number of bytes they requested.
# Indeed, this was the motivating reason for the complexity of this script;
# normally I would have used SimpleHTTPRequestHandler unmodified.
#
# Don't breathe too hard or this script might break.
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import http.server
# import time

FILE_SIZE = 1*1024*1024*1024  # 1 GiB


def _get_resp_size_from_range(range_str):
    assert range_str.startswith('bytes=')
    range_str = range_str[len('bytes='):]
    start_byte, end_byte = range_str.split('-')
    return int(end_byte) - int(start_byte) + 1


class MyHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    protocol_version = 'HTTP/1.1'

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    def send_head(self, length):
        self.send_response(200)
        self.send_header('Content-Type', 'application/octet-stream')
        self.send_header('Content-Length', length)
        # self.send_header('Last-Modified', self.date_time_string(time.time()))
        self.end_headers()

    def do_GET(self):
        range_hdr = self.headers['Range']
        if not range_hdr:
            num_bytes = FILE_SIZE
        else:
            assert range_hdr.startswith('bytes=')
            num_bytes = _get_resp_size_from_range(range_hdr)
        self.send_head(num_bytes)
        self.wfile.write(b'A' * num_bytes)

    def do_HEAD(self):
        self.send_head(FILE_SIZE)


def main(args):
    addr = ('', args.port)
    print('Listening on', addr)
    httpd = http.server.HTTPServer(addr, MyHTTPRequestHandler)
    httpd.serve_forever()


if __name__ == '__main__':
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-p', '--port', default=8000, type=int, help='Port on which to listen')
    args = parser.parse_args()
    try:
        exit(main(args))
    except KeyboardInterrupt:
        pass



"""Integration tests for requests."""
import uuid

import requests

from sbws import settings
from sbws.util import requests as requests_utils


def test_make_session(conf, persistent_launch_tor, dests):
    uuid_str = str(uuid.uuid4())
    settings.init_http_headers(
        conf.get("scanner", "nickname"),
        uuid_str,
        str(persistent_launch_tor.get_version()),
    )
    session = requests_utils.make_session(
        persistent_launch_tor, conf.getfloat("general", "http_timeout")
    )
    assert session._timeout == conf.getfloat("general", "http_timeout")

    # Because there is not an stream attached to a circuit, this will timeout.
    response = None
    try:
        response = session.get(dests.next().url, verify=False)
    except requests.exceptions.ConnectTimeout:
        pass
    assert response is None



from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

import threading  # noqa

from . import globals  # noqa


class Settings:
    """Singleton settings for all the packages.
    This way change settings can be seen by all the packages that import it.

    It lives in ``__init__.py`` to leave open the possibility of having a
    ``settings.py`` module for user settings.

    .. note:: After refactoring, globals should only have constants.
      Any other variable that needs to be modified when initializing
      should be initialized here.

    """

    def __init__(self):
        # update this dict from globals (but only for ALL_CAPS settings)
        for setting in dir(globals):
            if setting.isupper():
                setattr(self, setting, getattr(globals, setting))
        self.end_event = threading.Event()

    def init_http_headers(self, nickname, uuid, tor_version):
        self.HTTP_HEADERS["Tor-Bandwidth-Scanner-Nickname"] = nickname
        self.HTTP_HEADERS["Tor-Bandwidth-Scanner-UUID"] = uuid
        self.HTTP_HEADERS["User-Agent"] += tor_version

    def set_end_event(self):
        self.end_event.set()


settings = Settings()  # noqa



import logging
import os
import platform

from requests.__version__ import __version__ as requests_version
from stem import __version__ as stem_version

import sbws.core.cleanup
import sbws.core.flowctrl2
import sbws.core.generate
import sbws.core.scanner
import sbws.core.stats
from sbws import __version__ as version
from sbws.util.config import configure_logging, get_config, validate_config
from sbws.util.fs import sbws_required_disk_space
from sbws.util.parser import create_parser

log = logging.getLogger(__name__)


def _ensure_dirs(conf):
    log.debug("Ensuring all dirs exists.")
    # it is not needed to check sbws_home dir, since the following
    # will create parent dirs too (in case they don't exist)
    os.makedirs(conf.getpath("paths", "datadir"), exist_ok=True)
    os.makedirs(conf.getpath("paths", "v3bw_dname"), exist_ok=True)
    os.makedirs(conf.getpath("paths", "log_dname"), exist_ok=True)


def _adjust_log_level(args, conf):
    if not args.log_level:
        return
    conf["logger_sbws"]["level"] = args.log_level


def _get_startup_line():
    py_ver = platform.python_version()
    py_plat = platform.platform()
    return "sbws %s with python %s on %s, stem %s, and requests %s" % (
        version,
        py_ver,
        py_plat,
        stem_version,
        requests_version,
    )


def main():
    parser = create_parser()
    args = parser.parse_args()
    conf = get_config(args)
    _adjust_log_level(args, conf)
    conf_valid, conf_errors = validate_config(conf)
    if not conf_valid:
        for e in conf_errors:
            log.critical(e)
        exit(1)
    # Create directories after the home have been obtained from the config.
    _ensure_dirs(conf)
    configure_logging(args, conf)
    parser.description = sbws_required_disk_space(conf)
    def_args = [args, conf]
    def_kwargs = {}
    known_commands = {
        "cleanup": {
            "f": sbws.core.cleanup.main,
            "a": def_args,
            "kw": def_kwargs,
        },
        "scanner": {
            "f": sbws.core.scanner.main,
            "a": def_args,
            "kw": def_kwargs,
        },
        "generate": {
            "f": sbws.core.generate.main,
            "a": def_args,
            "kw": def_kwargs,
        },
        "stats": {"f": sbws.core.stats.main, "a": def_args, "kw": def_kwargs},
        "flowctrl2": {
            "f": sbws.core.flowctrl2.main,
            "a": def_args,
            "kw": def_kwargs,
        },
    }
    try:
        if args.command not in known_commands:
            parser.print_help()
        else:
            log.info(_get_startup_line())
            comm = known_commands[args.command]
            exit(comm["f"](*comm["a"], **comm["kw"]))
    except KeyboardInterrupt:
        print("")



def incrementing_time(start=2000, increment=1):
    while True:
        yield start
        start += increment


def monotonic_time(start=2000):
    return incrementing_time(start, increment=0.000001)


def static_time(value):
    while True:
        yield value



import logging
import os
import platform
from collections import OrderedDict

from requests import __version__ as requests_version
from stem import __version__ as stem_version

from sbws import __version__

log = logging.getLogger(__name__)

RESULT_VERSION = 4
WIRE_VERSION = 1
SPEC_VERSION = "1.5.0"

# This is a dictionary of torrc options we always want to set when launching
# Tor and that do not depend on any runtime configuration
# Options that are known at runtime (from configuration file) are added
# in utils/stem.py launch_tor
TORRC_STARTING_POINT = {
    # We will find out via the ControlPort and not setting something static
    # means a lower chance of conflict
    "SocksPort": "auto",
    # Easier than password authentication
    "CookieAuthentication": "1",
    # To avoid path bias warnings
    "UseEntryGuards": "0",
    # Because we need things from full server descriptors (namely for now: the
    # bandwidth line)
    "UseMicrodescriptors": "0",
    # useful logging options for clients that don't care about anonymity
    "SafeLogging": "0",
    "LogTimeGranularity": "1",
    "ProtocolWarnings": "1",
    # To be able to respond to MaxAdvertisedBandwidth as soon as possible.
    # If ``FetchDirInfoExtraEarly` is set, but not
    # `FetchDirInfoEarly`, Tor will throw this error:
    # `FetchDirInfoExtraEarly requires that you also set FetchDirInfoEarly`
    "FetchDirInfoEarly": "1",
    "FetchDirInfoExtraEarly": "1",
    # To make Tor keep fetching descriptors, even when idle.
    "FetchUselessDescriptors": "1",
    # Things needed to make circuits fail a little faster. We get the
    # circuit_timeout as a string instead of an int on purpose: stem only
    # accepts strings.
    "LearnCircuitBuildTimeout": "0",
}
# Options that need to be set at runtime.
TORRC_RUNTIME_OPTIONS = {
    # The scanner builds the circuits to download the data itself,
    # so do not let Tor to build them.
    "__DisablePredictedCircuits": "1",
    # The scanner attach the streams to the circuit itself,
    # so do not let Tor to attach them.
    "__LeaveStreamsUnattached": "1",
}
# Options that can be set at runtime and can fail with some Tor versions
# The ones that fail will be ignored..
TORRC_OPTIONS_CAN_FAIL = OrderedDict(
    {
        # Since currently scanner anonymity is not the goal, ConnectionPadding
        # is disable to do not send extra traffic
        "ConnectionPadding": "0",
    }
)

PKG_DIR = os.path.abspath(os.path.dirname(__file__))
DEFAULT_CONFIG_PATH = os.path.join(PKG_DIR, "config.default.ini")
DEFAULT_LOG_CONFIG_PATH = os.path.join(PKG_DIR, "config.log.default.ini")
USER_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".sbws.ini")
SUPERVISED_USER_CONFIG_PATH = "/etc/sbws/sbws.ini"
SUPERVISED_RUN_DPATH = "/run/sbws/tor"

SOCKET_TIMEOUT = 60  # seconds

SBWS_SCALE_CONSTANT = 7500
TORFLOW_SCALING = 1
SBWS_SCALING = 2
TORFLOW_BW_MARGIN = 0.05
TORFLOW_OBS_LAST = 0
TORFLOW_OBS_MEAN = 1
TORFLOW_OBS_DECAYING = 3
TORFLOW_ROUND_DIG = 3
PROP276_ROUND_DIG = 2
# Number of seconds the measurements for a relay have to be distant from each
# other, otherwise the relay would be excluded from the relays to vote on.
# Ideally, this should be 86400 seconds (1 day).
# To have sbws vote on approximately the same number of relays as Torflow,
# leave it as None, to not exclude measurements.
DAY_SECS = None
# Minimum number of measurements for a relay to be included as a relay to vote
# on.
# Ideally, this should be 2.
# As the constant before, leave it as 1 to not exclude measurements.
NUM_MIN_RESULTS = 1
MIN_REPORT = 60
# Maximum difference between the total consensus bandwidth and the total in
# in the bandwidth lines in percentage
MAX_BW_DIFF_PERC = 50

# With the new KeyValues in #29591, the lines are greater than 510
# Tor already accept lines of any size, but leaving the limit anyway.
BW_LINE_SIZE = 1022

# RelayList, ResultDump
# For how many seconds in the past the relays and measurements data is keep/
# considered valid.
# This is currently set by default in config.default.ini as ``data_period``,
# and used in ResultDump.
# In a future refactor, constants in config.default.ini should be moved here,
# or calculated in settings, so that there's no need to pass the configuration
# to all the functions.
MEASUREMENTS_PERIOD = 5 * 24 * 60 * 60

# #40017: To make sbws behave similar to Torflow, the number of raw past
# measurements used when generating the Bandwidth File has to be 28, not 5.
# Note that this is different from the number of raw past measurements used
# when measuring, which are used for the monitoring values and storing json.
GENERATE_PERIOD = 28 * 24 * 60 * 60

# Metadata to send in every requests, so that data servers can know which
# scanners are using them.
# In Requests these keys are case insensitive.
HTTP_HEADERS = {
    # This would be ignored if changing to HTTP/2
    "Connection": "keep-alive",
    # Needs to get Tor version from the controller
    "User-Agent": "sbws/{} ({}) Python/{} Requests/{} Stem/{} Tor/".format(
        __version__,
        platform.platform(),
        platform.python_version(),
        requests_version,
        stem_version,
    ),
    # Organization defined names (:rfc:`7239`)
    # Needs to get the nickname from the user config file.
    "Tor-Bandwidth-Scanner-Nickname": "{}",
    "Tor-Bandwidth-Scanner-UUID": "{}",
    # In case of including IP address.
    # 'Forwarded': 'for={}'  # IPv6 part, if there's
}
# In the case of having ipv6 it's concatenated to forwarder.
IPV6_FORWARDED = ', for="[{}]"'

HTTP_GET_HEADERS = {
    "Range": "{}",
    "Accept-Encoding": "identity",
}
DESTINATION_VERIFY_CERTIFICATE = True
# This number might need adjusted depending on the percentage of circuits and
# HTTP requests failures.

# Number of attempts to use a destination, that are stored, in order to decide
# whether the destination is functional or not.
NUM_DESTINATION_ATTEMPTS_STORED = 10
# Time to wait before trying again a destination that wasn't functional.
# Because intermitent failures with CDN destinations, start trying again
# after 5 min.
DELTA_SECONDS_RETRY_DESTINATION = 60 * 5
# No matter what, do not increase the wait time between destination reties
# past this value.
MAX_SECONDS_RETRY_DESTINATION = 60 * 60 * 3
# Number of consecutive times a destination can fail before considering it
# not functional.
MAX_NUM_DESTINATION_FAILURES = 3
# By which factor to multiply DELTA_SECONDS_RETRY_DESTINATION when the
# destination fail again.
FACTOR_INCREMENT_DESTINATION_RETRY = 2

# Constants to check health KeyValues in the bandwidth file
PERIOD_DAYS = int(MEASUREMENTS_PERIOD / (24 * 60 * 60))
MAX_RECENT_CONSENSUS_COUNT = PERIOD_DAYS * 24  # 120
# XXX: This was only defined in `config.default.ini`, it should be read from
# here.
FRACTION_RELAYS = 0.05
# A priority list currently takes more than 3h, ideally it should only take 1h.
MIN_HOURS_PRIORITY_LIST = 1
# As of 2020, there're less than 7000 relays.
MAX_RELAYS = 8000
# 120
MAX_RECENT_PRIORITY_LIST_COUNT = int(
    PERIOD_DAYS * 24 / MIN_HOURS_PRIORITY_LIST
)
MAX_RELAYS_PER_PRIORITY_LIST = int(MAX_RELAYS * FRACTION_RELAYS)  # 400
# 48000
MAX_RECENT_PRIORITY_RELAY_COUNT = (
    MAX_RECENT_PRIORITY_LIST_COUNT * MAX_RELAYS_PER_PRIORITY_LIST
)


def fail_hard(*a, **kw):
    """Log something ... and then exit as fast as possible"""
    log.critical(*a, **kw)
    exit(1)


def touch_file(fname, times=None):
    """
    If **fname** exists, update its last access and modified times to now. If
    **fname** does not exist, create it. If **times** are specified, pass them
    to os.utime for use.

    :param str fname: Name of file to update or create
    :param tuple times: 2-tuple of floats for access time and modified time
        respectively
    """
    log.debug("Touching %s", fname)
    with open(fname, "a") as fd:
        os.utime(fd.fileno(), times=times)



""" Measure the relays. """
import concurrent.futures
import logging
import os
import queue
import random
import signal
import sys
import threading
import time
import traceback
import uuid
from argparse import ArgumentDefaultsHelpFormatter

import sbws.util.requests as requests_utils
import sbws.util.stem as stem_utils
from sbws.globals import HTTP_GET_HEADERS, SOCKET_TIMEOUT, fail_hard

from .. import settings
from ..lib.circuitbuilder import GapsCircuitBuilder as CB
from ..lib.destination import (
    DestinationList,
    connect_to_destination_over_circuit,
)
from ..lib.heartbeat import Heartbeat
from ..lib.relaylist import RelayList
from ..lib.relayprioritizer import RelayPrioritizer
from ..lib.resultdump import (
    ResultDump,
    ResultError,
    ResultErrorCircuit,
    ResultErrorDestination,
    ResultErrorSecondRelay,
    ResultErrorStream,
    ResultSuccess,
)
from ..util.state import State
from ..util.timestamp import now_isodt_str

rng = random.SystemRandom()
log = logging.getLogger(__name__)
# Declare the objects that manage the threads global so that sbws can exit
# gracefully at any time.
rd = None
controller = None

FILLUP_TICKET_MSG = """Something went wrong.
Please create an issue at
https://gitlab.torproject.org/tpo/network-health/sbws/-/issues with this
traceback."""


def stop_threads(signal, frame, exit_code=0):
    global rd
    log.debug("Stopping sbws.")
    # Avoid new threads to start.
    settings.set_end_event()
    # Stop ResultDump thread
    rd.thread.join()
    # Stop Tor thread
    controller.close()
    sys.exit(exit_code)


signal.signal(signal.SIGTERM, stop_threads)


def dumpstacks():
    log.critical(FILLUP_TICKET_MSG)
    thread_id2name = dict([(t.ident, t.name) for t in threading.enumerate()])
    for thread_id, stack in sys._current_frames().items():
        log.critical(
            "Thread: %s(%d)", thread_id2name.get(thread_id, ""), thread_id
        )
        log.critical("Traceback: %s", "".join(traceback.format_stack(stack)))


def sigint_handler():
    import pdb

    pdb.set_trace()


signal.signal(signal.SIGINT, sigint_handler)


def timed_recv_from_server(session, dest, byte_range):
    """Request the **byte_range** from the URL at **dest**. If successful,
    return True and the time it took to download. Otherwise return False and an
    exception."""

    start_time = time.time()
    HTTP_GET_HEADERS["Range"] = byte_range
    # - response.elapsed "measures the time taken between sending the first
    #   byte of the request and finishing parsing the headers.
    #   It is therefore unaffected by consuming the response content"
    #   If this mean that the content has arrived, elapsed could be used to
    #   know the time it took.
    try:
        # headers are merged with the session ones, not overwritten.
        session.get(dest.url, headers=HTTP_GET_HEADERS, verify=dest.verify)
    # All `requests` exceptions could be caught with
    # `requests.exceptions.RequestException`, but it seems that `requests`
    # does not catch all the ssl exceptions and urllib3 doesn't seem to have
    # a base exception class.
    except Exception as e:
        log.debug(e)
        return False, e
    end_time = time.time()
    return True, end_time - start_time


def get_random_range_string(content_length, size):
    """
    Return a random range of bytes of length **size**. **content_length** is
    the size of the file we will be requesting a range of bytes from.

    For example, for content_length of 100 and size 10, this function will
    return one of the following: '0-9', '1-10', '2-11', [...] '89-98', '90-99'
    """
    assert size <= content_length
    # start can be anywhere in the content_length as long as it is **size**
    # bytes away from the end or more. Because range is [start, end) (doesn't
    # include the end value), add 1 to the end.
    start = rng.choice(range(0, content_length - size + 1))
    # Unlike range, the byte range in an http header is [start, end] (does
    # include the end value), so we subtract one
    end = start + size - 1
    # start and end are indexes, while content_length is a length, therefore,
    # the largest index end should ever be should be less than the total length
    # of the content. For example, if content_length is 10, end could be
    # anywhere from 0 to 9.
    assert end < content_length
    return "bytes={}-{}".format(start, end)


def measure_rtt_to_server(session, conf, dest, content_length):
    """Make multiple end-to-end RTT measurements by making small HTTP requests
    over a circuit + stream that should already exist, persist, and not need
    rebuilding. If something goes wrong and not all of the RTT measurements can
    be made, return None. Otherwise return a list of the RTTs (in seconds).

    :returns tuple: results or None if the if the measurement fail.
        None or exception if the measurement fail.

    """
    rtts = []
    size = conf.getint("scanner", "min_download_size")
    for _ in range(0, conf.getint("scanner", "num_rtts")):
        log.debug("Measuring RTT to %s", dest.url)
        random_range = get_random_range_string(content_length, size)
        success, data = timed_recv_from_server(session, dest, random_range)
        if not success:
            # data is an exception
            log.debug(
                "While measuring the RTT to %s we hit an exception "
                "(does the webserver support Range requests?): %s",
                dest.url,
                data,
            )
            return None, data
        assert success
        # data is an RTT
        assert isinstance(data, float) or isinstance(data, int)
        rtts.append(data)
    return rtts, None


def measure_bandwidth_to_server(session, conf, dest, content_length):
    """
    :returns tuple: results or None if the if the measurement fail.
        None or exception if the measurement fail.

    """
    results = []
    num_downloads = conf.getint("scanner", "num_downloads")
    expected_amount = conf.getint("scanner", "initial_read_request")
    min_dl = conf.getint("scanner", "min_download_size")
    max_dl = conf.getint("scanner", "max_download_size")
    download_times = {
        "toofast": conf.getfloat("scanner", "download_toofast"),
        "min": conf.getfloat("scanner", "download_min"),
        "target": conf.getfloat("scanner", "download_target"),
        "max": conf.getfloat("scanner", "download_max"),
    }
    while len(results) < num_downloads and not settings.end_event.is_set():
        assert expected_amount >= min_dl
        assert expected_amount <= max_dl
        random_range = get_random_range_string(content_length, expected_amount)
        success, data = timed_recv_from_server(session, dest, random_range)
        if not success:
            # data is an exception
            log.debug(
                "While measuring the bandwidth to %s we hit an "
                "exception (does the webserver support Range "
                "requests?): %s",
                dest.url,
                data,
            )
            return None, data
        assert success
        # data is a download time
        assert isinstance(data, float) or isinstance(data, int)
        if _should_keep_result(
            expected_amount == max_dl, data, download_times
        ):
            results.append({"duration": data, "amount": expected_amount})
        expected_amount = _next_expected_amount(
            expected_amount, data, download_times, min_dl, max_dl
        )
    return results, None


def _pick_ideal_second_hop(relay, rl, relay_as_entry, candidates):
    """
    Sbws builds two hop circuits. Given the **relay** to measure, pick a helper
    relay from **candidates** with minimum bandwidth depending on whether
    **relay_as_entry**.
    """
    # 40041: Instead of using exits that can exit to all IPs, to ensure that
    # they can make requests to the Web servers, try with the exits that
    # allow some IPs, since there're more.
    # In the case that a concrete exit can't exit to the Web server, it is not
    # a problem since the relay will be measured in the next loop with other
    # random exit.

    # While not all exits implement congestion control, the min bw might not
    # correspond to the subset that implement it.
    min_relay_bw = rl.exit_min_bw() if relay_as_entry else rl.non_exit_min_bw()
    log.debug(
        "Picking a 2nd hop to measure %s from %d choices. relay_as_entry=%s",
        relay.nickname,
        len(candidates),
        relay_as_entry,
    )
    for min_bw_factor in [2, 1.75, 1.5, 1.25, 1]:
        min_bw = relay.consensus_bandwidth * min_bw_factor
        # We might have a really slow/new relay. Try to measure it properly by
        # using only relays with or above our calculated min_relay_bw (see:
        # _calculate_min_bw_second_hop() in relaylist.py).
        if min_bw < min_relay_bw:
            min_bw = min_relay_bw
        new_candidates = stem_utils.only_relays_with_bandwidth(
            candidates, min_bw=min_bw
        )
        if len(new_candidates) > 0:
            chosen = rng.choice(new_candidates)
            log.debug(
                "Found %d candidate 2nd hops with at least %sx the bandwidth "
                "of %s. Returning %s (bw=%s).",
                len(new_candidates),
                min_bw_factor,
                relay.nickname,
                chosen.nickname,
                chosen.consensus_bandwidth,
            )
            return chosen
    candidates = sorted(
        candidates, key=lambda r: r.consensus_bandwidth, reverse=True
    )
    chosen = candidates[0]
    log.debug(
        "Didn't find any 2nd hops at least as fast as %s (bw=%s). It's "
        "probably really fast. Returning %s (bw=%s), the fastest "
        "candidate we have.",
        relay.nickname,
        relay.consensus_bandwidth,
        chosen.nickname,
        chosen.consensus_bandwidth,
    )
    return chosen


def error_no_helper(relay, dest, our_nick=""):
    reason = "Unable to select a second relay"
    log.debug(
        reason + " to help measure %s (%s)", relay.fingerprint, relay.nickname
    )
    return [
        ResultErrorSecondRelay(relay, [], dest.url, our_nick, msg=reason),
    ]


def create_path_relay(relay, dest, rl, relay_as_entry, candidates):
    # the helper `relay_as_entry` arg,
    # is True when the relay is the entry (helper has to be exit)
    # and False when the relay is not the entry, ie. is the exit (helper does
    # not have to be an exit)

    if not candidates:
        return error_no_helper(relay, dest)

    helper = _pick_ideal_second_hop(relay, rl, relay_as_entry, candidates)

    if not helper:
        return error_no_helper(relay, dest)
    if relay_as_entry:
        circ_fps = [relay.fingerprint, helper.fingerprint]
        nicknames = [relay.nickname, helper.nickname]
        exit_policy = helper.exit_policy
    else:
        circ_fps = [helper.fingerprint, relay.fingerprint]
        nicknames = [helper.nickname, relay.nickname]
        exit_policy = relay.exit_policy
    return circ_fps, nicknames, exit_policy


def error_no_circuit(circ_fps, nicknames, reason, relay, dest, our_nick):
    log.debug(
        "Could not build circuit with path %s (%s): %s ",
        circ_fps,
        nicknames,
        reason,
    )
    return [
        ResultErrorCircuit(relay, circ_fps, dest.url, our_nick, msg=reason),
    ]


def use_relay_as_entry(relay, rl, dest):
    """Return whether to use the relay as entry or not.

    :rtype: bool

    """
    log.debug("Deciding whether to use the relay as entry or not.")
    relay_as_entry = True
    if rl.is_consensus_cc_alg_2:
        log.debug("Congestion control enabled.")
        if rl.is_consensus_bwscanner_cc_gte_1:
            log.debug("Use congestion control.")
            if (
                relay.is_exit_not_bad_allowing_port(dest.port)
                and relay.has_2_in_flowctrl
            ):
                log.debug("Relay to measure is exit.")
                log.debug("Exit has 2 in FlowCtrl.")
                log.debug("Use relay as exit. Choose non-exits.")
                relay_as_entry = False
            else:  # no exit or no 2 in FlowCtrl
                log.debug("Relay is not exit or has NOT 2 in FlowCtrl.")
                log.debug(
                    "Use relay as entry."
                    "Choose an exit that does have 2 in FlowCtrl"
                )
        else:  # bwscanner_cc != 1
            log.debug("Do not use congestion control.")
            if (
                relay.is_exit_not_bad_allowing_port(dest.port)
                and not relay.has_2_in_flowctrl
            ):
                log.debug("Relay to measure is exit.")
                log.debug("Exit has NOT 2 in FlowCtrl.")
                log.debug("Use relay as exit. Choose non-exits.")
                relay_as_entry = False
            else:  # no exit or 2 in FlowCtrl
                log.debug("Relay is not exit or it has 2 in FlowCtrl.")
                log.debug(
                    "Use relay as entry."
                    "Choose an exit that does NOT have 2 in FlowCtrl"
                )
    else:  # cc_alg!=2
        log.debug("Congestion control disabled.")
        if relay.is_exit_not_bad_allowing_port(dest.port):
            log.debug("Relay to measure is exit.")
            log.debug("Use relay as exit. Choose non-exits.")
            relay_as_entry = False
        else:
            log.debug("Relay to measure is NOT exit.")
            log.debug("Use relay as entry. Choose an exit.")
    return relay_as_entry


def select_helper_candidates(relay, rl, dest, relay_as_entry=True):
    """Return helper candidates list.

    :rtype: list

    """
    if rl.is_consensus_cc_alg_2:
        log.debug("Congestion control enabled.")
        if rl.is_consensus_bwscanner_cc_gte_1:
            log.debug("Use congestion control.")
            if (
                relay.is_exit_not_bad_allowing_port(dest.port)
                and relay.has_2_in_flowctrl
                and not relay_as_entry
            ):
                log.debug("Relay to measure is exit.")
                log.debug("Exit has 2 in FlowCtrl.")
                log.debug("Use relay as exit. Choose non-exits.")
                candidates = rl.non_exits
            else:  # no exit or no 2 in FlowCtrl
                log.debug("Relay is not exit or has NOT 2 in FlowCtrl.")
                log.debug(
                    "Use relay as entry."
                    "Choose an exit that does have 2 in FlowCtrl"
                )
                candidates = rl.exits_with_2_in_flowctrl(dest.port)
        else:  # bwscanner_cc != 1
            log.debug("Do not use congestion control.")
            if (
                relay.is_exit_not_bad_allowing_port(dest.port)
                and not relay.has_2_in_flowctrl
                and not relay_as_entry
            ):
                log.debug("Relay to measure is exit.")
                log.debug("Exit has NOT 2 in FlowCtrl.")
                log.debug("Use relay as exit. Choose non-exits.")
                candidates = rl.non_exits
            else:  # no exit or 2 in FlowCtrl
                log.debug("Relay is not exit or it has 2 in FlowCtrl.")
                log.debug(
                    "Use relay as entry."
                    "Choose an exit that does NOT have 2 in FlowCtrl"
                )
                candidates = rl.exits_without_2_in_flowctrl(dest.port)
    else:  # cc_alg!=2
        log.debug("Congestion control disabled.")
        if (
            relay.is_exit_not_bad_allowing_port(dest.port)
            and not relay_as_entry
        ):
            log.debug("Relay to measure is exit.")
            log.debug("Use relay as exit. Choose non-exits.")
            candidates = rl.non_exits
        else:
            log.debug("Relay to measure is NOT exit.")
            log.debug("Use relay as entry. Choose an exit.")
            candidates = rl.exits_not_bad_allowing_port(dest.port)

    # In the case the relay is measured as an exit, the entry helper could be
    # an exit too
    # (#40041), so ensure the helper is not the same as the entry, likely to
    # happen in a test network.
    if not relay_as_entry:  # relay to measure as exit
        candidates = [
            c for c in candidates if c.fingerprint != relay.fingerprint
        ]
    return candidates


def measure_relay(args, conf, destinations, cb, rl, relay):
    """
    Select a Web server, a relay to build the circuit,
    build the circuit and measure the bandwidth of the given relay.

    :return Result: a measurement Result object

    """
    log.debug("Measuring %s %s", relay.nickname, relay.fingerprint)
    our_nick = conf["scanner"]["nickname"]
    s = requests_utils.make_session(
        cb.controller, conf.getfloat("general", "http_timeout")
    )
    # Probably because the scanner is stopping.
    if s is None:
        if settings.end_event.is_set():
            return None
        else:
            # In future refactor this should be returned from the make_session
            reason = "Unable to get proxies."
            log.debug(
                reason + " to measure %s %s", relay.nickname, relay.fingerprint
            )
            return [
                ResultError(relay, [], "", our_nick, msg=reason),
            ]
    # Pick a destination
    dest = destinations.next()
    # When there're no any functional destinations.
    if not dest:
        # NOTE: When there're still functional destinations but only one of
        # them fail, the error will be included in `ResultErrorStream`.
        # Since this is being executed in a thread, the scanner can not
        # be stop here, but the `end_event` signal can be set so that the
        # main thread stop the scanner.
        # It might be useful to store the fact that the destinations fail,
        # so store here the error, and set the signal once the error is stored
        # (in `resultump`).
        log.critical(
            "There are not any functional destinations.\n"
            "It is recommended to set several destinations so that "
            "the scanner can continue if one fails."
        )
        reason = "No functional destinations"
        # Resultdump will set end_event after storing the error
        return [
            ResultErrorDestination(relay, [], "", our_nick, msg=reason),
        ]

    # Pick a relay to help us measure the given relay. If the given relay is an
    # exit, then pick a non-exit. Otherwise pick an exit.
    # From #40125, this condition is more complex.

    # Instead of ensuring that the relay can exit to all IPs, try first with
    # the relay as an exit, if it can exit to some IPs.

    # #40125 Check whether the relay will be used as entry and which obtain the
    # helper candidates.
    relay_as_entry = use_relay_as_entry(relay, rl, dest)
    candidates = select_helper_candidates(relay, rl, dest, relay_as_entry)
    r = create_path_relay(relay, dest, rl, relay_as_entry, candidates)

    # When `error_no_helper` is triggered because a helper is not found, what
    # can happen in test networks with very few relays, it returns a list with
    # the error.
    if len(r) == 1:
        return r
    circ_fps, nicknames, exit_policy = r

    # Build the circuit
    circ_id, reason = cb.build_circuit(circ_fps)

    # If the circuit failed to get created, bad luck, it will be created again
    # with other helper.
    # Here we won't have the case that an exit tried to build the circuit as
    # entry and failed (#40029), cause not checking that it can exit all IPs.
    if not circ_id:
        return error_no_circuit(
            circ_fps, nicknames, reason, relay, dest, our_nick
        )
    log.debug(
        "Built circuit with path %s (%s) to measure %s (%s)",
        circ_fps,
        nicknames,
        relay.fingerprint,
        relay.nickname,
    )
    # Make a connection to the destination
    is_usable, usable_data = connect_to_destination_over_circuit(
        dest, circ_id, s, cb.controller, dest._max_dl
    )

    # In the case that the relay was used as an exit, but could not exit
    # to the Web server, try again using it as entry, to avoid that it would
    # always fail when there's only one Web server.

    if not is_usable and not relay_as_entry:
        log.debug(
            "Exit %s (%s) that can't exit all ips, with exit policy %s, failed"
            " to connect to %s via circuit %s (%s). Reason: %s. Trying again "
            "with it as entry.",
            relay.fingerprint,
            relay.nickname,
            exit_policy,
            dest.url,
            circ_fps,
            nicknames,
            usable_data,
        )
        relay_as_entry = True
        # select new candidates as exit
        candidates = select_helper_candidates(relay, rl, dest, relay_as_entry)
        r = create_path_relay(relay, dest, rl, relay_as_entry, candidates)
        if len(r) == 1:
            return r
        circ_fps, nicknames, exit_policy = r
        circ_id, reason = cb.build_circuit(circ_fps)
        if not circ_id:
            log.info(
                "Exit %s (%s) that can't exit all ips, failed to create "
                " circuit as entry: %s (%s).",
                relay.fingerprint,
                relay.nickname,
                circ_fps,
                nicknames,
            )
            return error_no_circuit(
                circ_fps, nicknames, reason, relay, dest, our_nick
            )

        log.debug(
            "Built circuit with path %s (%s) to measure %s (%s)",
            circ_fps,
            nicknames,
            relay.fingerprint,
            relay.nickname,
        )
        is_usable, usable_data = connect_to_destination_over_circuit(
            dest, circ_id, s, cb.controller, dest._max_dl
        )
    if not is_usable:
        log.debug(
            "Failed to connect to %s to measure %s (%s) via circuit "
            "%s (%s). Exit policy: %s. Reason: %s.",
            dest.url,
            relay.fingerprint,
            relay.nickname,
            circ_fps,
            nicknames,
            exit_policy,
            usable_data,
        )
        cb.close_circuit(circ_id)
        return [
            ResultErrorStream(
                relay, circ_fps, dest.url, our_nick, msg=usable_data
            ),
        ]
    assert is_usable
    assert "content_length" in usable_data
    # FIRST: measure RTT
    rtts, reason = measure_rtt_to_server(
        s, conf, dest, usable_data["content_length"]
    )
    if rtts is None:
        log.debug(
            "Unable to measure RTT for %s (%s) to %s via circuit "
            "%s (%s): %s",
            relay.fingerprint,
            relay.nickname,
            dest.url,
            circ_fps,
            nicknames,
            reason,
        )
        cb.close_circuit(circ_id)
        return [
            ResultErrorStream(
                relay, circ_fps, dest.url, our_nick, msg=str(reason)
            ),
        ]
    # SECOND: measure bandwidth
    bw_results, reason = measure_bandwidth_to_server(
        s, conf, dest, usable_data["content_length"]
    )
    if bw_results is None:
        log.debug(
            "Failed to measure %s (%s) via circuit %s (%s) to %s. Exit"
            " policy: %s. Reason: %s.",
            relay.fingerprint,
            relay.nickname,
            circ_fps,
            nicknames,
            dest.url,
            exit_policy,
            reason,
        )
        cb.close_circuit(circ_id)
        return [
            ResultErrorStream(
                relay, circ_fps, dest.url, our_nick, msg=str(reason)
            ),
        ]
    cb.close_circuit(circ_id)
    # Finally: store result
    log.debug(
        "Success measurement for %s (%s) via circuit %s (%s) to %s",
        relay.fingerprint,
        relay.nickname,
        circ_fps,
        nicknames,
        dest.url,
    )
    return [
        ResultSuccess(rtts, bw_results, relay, circ_fps, dest.url, our_nick),
    ]


def dispatch_worker_thread(*a, **kw):
    # If at the point where the relay is actually going to be measured there
    # are not any functional destinations or the `end_event` is set, do not
    # try to start measuring the relay, since it will fail anyway.
    try:
        # a[2] is the argument `destinations`
        functional_destinations = a[2].functional_destinations
    # In case the arguments or the method change, catch the possible exceptions
    # but ignore here that there are not destinations.
    except (IndexError, TypeError):
        log.debug("Wrong argument or attribute.")
        functional_destinations = True
    if not functional_destinations or settings.end_event.is_set():
        return None
    return measure_relay(*a, **kw)


def _should_keep_result(did_request_maximum, result_time, download_times):
    # In the normal case, we didn't ask for the maximum allowed amount. So we
    # should only allow ourselves to keep results that are between the min and
    # max allowed time
    msg = "Keeping measurement time {:.2f}".format(result_time)
    if (
        not did_request_maximum
        and result_time >= download_times["min"]
        and result_time < download_times["max"]
    ):
        log.debug(msg)
        return True
    # If we did request the maximum amount, we should keep the result as long
    # as it took less than the maximum amount of time
    if did_request_maximum and result_time < download_times["max"]:
        log.debug(msg)
        return True
    # In all other cases, return false
    log.debug(
        "Not keeping result time %f.%s",
        result_time,
        ""
        if not did_request_maximum
        else " We requested the maximum " "amount allowed.",
    )
    return False


def _next_expected_amount(
    expected_amount, result_time, download_times, min_dl, max_dl
):
    if result_time < download_times["toofast"]:
        # Way too fast, greatly increase the amount we ask for
        expected_amount = int(expected_amount * 5)
    elif (
        result_time < download_times["min"]
        or result_time >= download_times["max"]
    ):
        # As long as the result is between min/max, keep the expected amount
        # the same. Otherwise, adjust so we are aiming for the target amount.
        expected_amount = int(
            expected_amount * download_times["target"] / result_time
        )
    # Make sure we don't request too much or too little
    expected_amount = max(min_dl, expected_amount)
    expected_amount = min(max_dl, expected_amount)
    return expected_amount


def measurement_writer(result_dump, measurement):
    # Since result_dump thread is calling queue.get() every second,
    # the queue should be full for only 1 second.
    # This call blocks at maximum timeout seconds.
    try:
        result_dump.queue.put(measurement, timeout=3)
    except queue.Full:
        # The result would be lost, the scanner will continue working.
        log.warning(
            "The queue with measurements is full, when adding %s.\n"
            "It is possible that the thread that get them to "
            "write them to the disk (ResultDump.enter) is stalled.",
            measurement,
        )


def log_measurement_exception(target, exception):
    print("in result putter error")
    if settings.end_event.is_set():
        return
    # The only object that can be here if there is not any uncatched
    # exception is stem.SocketClosed when stopping sbws
    # An exception here means that the worker thread finished.
    log.warning(FILLUP_TICKET_MSG)
    # To print the traceback that happened in the thread, not here in
    # the main process.
    log.warning(
        "".join(
            traceback.format_exception(
                type(exception), exception, exception.__traceback__
            )
        )
    )
    log.debug(
        "".join(
            target.fingerprint,
            target.nickname,
            traceback.format_exception(
                type(exception), exception, exception.__traceback__
            ),
        )
    )


def main_loop(
    args,
    conf,
    controller,
    relay_list,
    circuit_builder,
    result_dump,
    relay_prioritizer,
    destinations,
):
    r"""Create the queue of future measurements for every relay to measure.

    It starts a loop that will be run while there is not and event signaling
    that sbws is stopping (because of SIGTERM or SIGINT).

    Then the ``ThreadPoolExecutor`` (executor) queues all the relays to
    measure in ``Future`` objects. These objects have an ``state``.

    The executor starts a new thread for every relay to measure, which runs
    ``measure_relay`` until there are ``max_pending_results`` threads.
    After that, it will reuse a thread that has finished for every relay to
    measure.

    Then ``process_completed_futures`` is call, to obtain the results in the
    completed ``future``\s.

    """
    log.info("Started the main loop to measure the relays.")
    hbeat = Heartbeat(conf.getpath("paths", "state_fname"))

    # Set the time to wait for a thread to finish as the half of an HTTP
    # request timeout.
    # Do not start a new loop if sbws is stopping.
    while not settings.end_event.is_set():
        log.debug("Starting a new measurement loop.")
        num_relays = 0
        loop_tstart = time.time()

        # Register relay fingerprints to the heartbeat module
        hbeat.register_consensus_fprs(relay_list.relays_fingerprints)
        # num_threads
        max_pending_results = conf.getint("scanner", "measurement_threads")
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_pending_results, thread_name_prefix="measurer"
        ) as executor:
            log.info("In the executor, queue all future measurements.")
            # With futures, there's no need for callback, what it was the
            # callback with multiprocessing library can be just a function
            # that gets executed when the future result is obtained.
            pending_results = {
                executor.submit(
                    dispatch_worker_thread,
                    args,
                    conf,
                    destinations,
                    circuit_builder,
                    relay_list,
                    target,
                ): target
                for target in relay_prioritizer.best_priority()
            }
            log.debug("Measurements queued.")
            # After the submitting all the targets to the executor, the pool
            # has queued all the relays and pending_results has the list of all
            # `Future`s.

            # Each target relay_recent_measurement_attempt is incremented in
            # `process_completed_futures` as well as hbeat measured
            # fingerprints.
            num_relays = len(pending_results)
            # Without a callback, it's needed to pass `result_dump` here to
            # call the function that writes the measurement when it's
            # finished.
            process_completed_futures(
                executor,
                hbeat,
                result_dump,
                pending_results,
            )
            wait_futures_completed(pending_results)

        # Print the heartbeat message
        hbeat.print_heartbeat_message()

        loop_tstop = time.time()
        loop_tdelta = (loop_tstop - loop_tstart) / 60
        # At this point, we know the relays that were queued to be
        # measured.
        log.debug(
            "Attempted to measure %s relays in %i minutes.",
            num_relays,
            loop_tdelta,
        )
        # In a testing network, exit after first loop
        if controller.get_conf("TestingTorNetwork") == "1":
            log.info("In a testing network, exiting after the first loop.")
            # Threads should be closed nicely in some refactor
            stop_threads(signal.SIGTERM, None)


def process_completed_futures(executor, hbeat, result_dump, pending_results):
    """Obtain the relays' measurements as they finish.

    For every ``Future`` measurements that gets completed, obtain the
    ``result`` and call ``measurement_writer``, which put the ``Result`` in
    ``ResultDump`` queue and complete immediately.

    ``ResultDump`` thread (started before and out of this function) will get
    the ``Result`` from the queue and write it to disk, so this doesn't block
    the measurement threads.

    If there was an exception not caught by ``measure_relay``, it will call
    instead ``log_measurement_exception``, which logs the error and complete
    immediately.

    """
    num_relays_to_measure = num_pending_results = len(pending_results)
    with executor:
        for future_measurement in concurrent.futures.as_completed(
            pending_results
        ):
            target = pending_results[future_measurement]
            # 40023, disable to decrease state.dat json lines
            # relay_list.increment_recent_measurement_attempt()
            target.increment_relay_recent_measurement_attempt()

            # Register this measurement to the heartbeat module
            hbeat.register_measured_fpr(target.fingerprint)
            log.debug(
                "Future measurement for target %s (%s) is done: %s",
                target.fingerprint,
                target.nickname,
                future_measurement.done(),
            )
            try:
                measurement = future_measurement.result()
            except Exception as e:
                log_measurement_exception(target, e)
                import psutil

                log.warning(psutil.Process(os.getpid()).memory_full_info())
                virtualMemoryInfo = psutil.virtual_memory()
                availableMemory = virtualMemoryInfo.available
                log.warning(
                    "Memory available %s MB.", availableMemory / 1024**2
                )
                dumpstacks()
            else:
                log.info("Measurement ready: %s" % (measurement))
                measurement_writer(result_dump, measurement)
            # `pending_results` has all the initial queued `Future`s,
            # they don't decrease as they get completed, but we know 1 has be
            # completed in each loop,
            num_pending_results -= 1
            log.info(
                "Pending measurements: %s out of %s: ",
                num_pending_results,
                num_relays_to_measure,
            )


def wait_futures_completed(pending_results):
    """Wait for last futures to finish, before starting new loop."""
    log.info("Wait for any remaining measurements.")
    done, not_done = concurrent.futures.wait(
        pending_results,
        timeout=SOCKET_TIMEOUT + 10,  # HTTP timeout is 10
        return_when=concurrent.futures.ALL_COMPLETED,
    )
    log.info("Completed futures: %s", len(done))
    # log.debug([f.__dict__ for f in done])
    cancelled = [f for f in done if f.cancelled()]
    if cancelled:
        log.warning("Cancelled futures: %s", len(cancelled))
        for f, t in cancelled:
            log.debug(t.fingerprint)
            dumpstacks()
    if not_done:
        log.warning("Not completed futures: %s", len(not_done))
        for f, t in not_done:
            log.debug(t.fingerprint)
            dumpstacks()


def run_speedtest(args, conf):
    """Initializes all the data and threads needed to measure the relays.

    It launches or connect to Tor in a thread.
    It initializes the list of relays seen in the Tor network.
    It starts a thread to read the previous measurements and wait for new
    measurements to write them to the disk.
    It initializes a class that will be used to order the relays depending
    on their measurements age.
    It initializes the list of destinations that will be used for the
    measurements.
    It initializes the thread pool that will launch the measurement threads.
    The pool starts 3 other threads that are not the measurement (worker)
    threads.
    Finally, it calls the function that will manage the measurement threads.

    """
    global rd, controller

    controller = stem_utils.launch_or_connect_to_tor(conf)

    # When there will be a refactor where conf is global, this can be removed
    # from here.
    state = State(conf.getpath("paths", "state_fname"))
    # XXX: tech-debt: create new function to obtain the controller and to
    # write the state, so that a unit test to check the state tor version can
    # be created
    # Store tor version whenever the scanner starts.
    state["tor_version"] = str(controller.get_version())
    # Call only once to initialize http_headers
    settings.init_http_headers(
        conf.get("scanner", "nickname"), state["uuid"], state["tor_version"]
    )
    # To do not have to pass args and conf to RelayList, pass an extra
    # argument with the data_period
    measurements_period = conf.getint("general", "data_period")
    rl = RelayList(args, conf, controller, measurements_period, state)
    cb = CB(args, conf, controller, rl)
    rd = ResultDump(args, conf)
    rp = RelayPrioritizer(args, conf, rl, rd)
    destinations, error_msg = DestinationList.from_config(
        conf, cb, rl, controller
    )
    if not destinations:
        fail_hard(error_msg)
    try:
        main_loop(args, conf, controller, rl, cb, rd, rp, destinations)
    except KeyboardInterrupt:
        log.info("Interrupted by the user.")
        dumpstacks()
    except Exception as e:
        log.exception(e)
        dumpstacks()


def gen_parser(sub):
    d = (
        "The scanner side of sbws. This should be run on a well-connected "
        "machine on the Internet with a healthy amount of spare bandwidth. "
        "This continuously builds circuits, measures relays, and dumps "
        "results into a datadir, commonly found in ~/.sbws"
    )
    sub.add_parser(
        "scanner", formatter_class=ArgumentDefaultsHelpFormatter, description=d
    )


def main(args, conf):
    if conf.getint("scanner", "measurement_threads") < 1:
        fail_hard("Number of measurement threads must be larger than 1")

    min_dl = conf.getint("scanner", "min_download_size")
    max_dl = conf.getint("scanner", "max_download_size")
    if max_dl < min_dl:
        fail_hard(
            "Max download size %d cannot be smaller than min %d",
            max_dl,
            min_dl,
        )

    os.makedirs(conf.getpath("paths", "datadir"), exist_ok=True)

    state = State(conf.getpath("paths", "state_fname"))
    state["scanner_started"] = now_isodt_str()
    # Generate an unique identifier for each scanner
    if "uuid" not in state:
        state["uuid"] = str(uuid.uuid4())

    run_speedtest(args, conf)



"""timestamps.py unit tests."""

from datetime import datetime, timedelta

from sbws.util.state import State
from sbws.util.timestamps import DateTimeIntSeq, DateTimeSeq


def test_update_datetime_seq(conf):
    now = datetime.utcnow().replace(microsecond=0)
    state = State(conf["paths"]["state_fpath"])
    # Create a list of 6 datetimes that started 6 days in the past.
    dts = [now - timedelta(days=x) for x in range(6, 0, -1)]
    dt_seq = DateTimeSeq(
        dts, state=state, state_key="recent_measurement_attempt"
    )
    new_dts = dt_seq.update()
    # The updated list will not contain the 2 first (left) datetimes and it
    # will have one last timestamp (right).
    assert new_dts[:-1] == dts[2:]
    assert 5 == state.count("recent_measurement_attempt")
    assert 5 == len(dt_seq)


def test_update_datetime_int_seq(conf):
    now = datetime.utcnow().replace(microsecond=0)
    state = State(conf["paths"]["state_fpath"])
    # Create a list of 6 datetimes that started 6 days in the past.
    dts = [[now - timedelta(days=x), 2] for x in range(6, 0, -1)]
    dt_seq = DateTimeIntSeq(
        dts, state=state, state_key="recent_measurement_attempt"
    )
    new_dts = dt_seq.update()
    # The updated list will not contain the 2 first (left) tuples and it
    # will have one last tuple (right).
    # The last tuple have 0 as the integer, instead of 2, so the count will be
    # 2 * 4 = 8
    assert new_dts[:-1] == dts[2:]
    assert 8 == state.count("recent_measurement_attempt")
    # And `len` should return the same.
    assert 8 == len(dt_seq)


def test_last_datetime_seq(conf):
    dt_seq = DateTimeSeq([])
    new_dts = dt_seq.last()
    assert new_dts == datetime.utcnow().replace(microsecond=0) - timedelta(
        hours=1
    )


def test_create_list_datetime_seq(conf):
    now = datetime.utcnow().replace(microsecond=0)
    # Create a list of 6 datetimes that started 6 days in the past.
    dts = [now - timedelta(days=x) for x in range(6, 0, -1)]
    dt_seq = DateTimeSeq(dts)
    new_dts = dt_seq.list()
    assert type(new_dts) == list


def test_create_list_dt_int_seq(conf):
    now = datetime.utcnow().replace(microsecond=0)
    # Create a list of 6 datetimes that started 6 days in the past.
    dts = [now - timedelta(days=x) for x in range(6, 0, -1)]
    dt_int_seq = DateTimeIntSeq(dts)
    new_dt_int_seq = dt_int_seq.list()

    assert type(new_dt_int_seq) == list


def test_last_dt_int_seq(conf):
    dt_seq = DateTimeIntSeq([])
    assert (
        datetime.utcnow().replace(microsecond=0) - timedelta(hours=1)
        == dt_seq.last()
    )



"""pytest configuration for integration tests."""
import argparse
import os

import pytest
from stem.version import Version

from sbws.lib.circuitbuilder import GapsCircuitBuilder as CB
from sbws.lib.destination import DestinationList
from sbws.lib.relaylist import RelayList
from sbws.util.config import _get_default_config
from sbws.util.stem import launch_or_connect_to_tor


class _PseudoArguments(argparse.Namespace):

    """Just enough of the argparse.Namespace (what you get when you do
    args = parser.parse_args()) to make get_config() happy

    >>> args = _PseudoArguments(directory='/home/matt/.sbws')
    >>> args.directory
    '/home/matt/.sbws'

    """

    def __init__(self, **kw):
        for key in kw:
            setattr(self, key, kw[key])


@pytest.fixture(scope="session")
def tmpdir(tmpdir_factory, request):
    """Create a tmp dir for the tests"""
    base = str(hash(request.node.nodeid))[:3]
    bn = tmpdir_factory.mktemp(base)
    return bn


@pytest.fixture(scope="session")
def sbwshome_empty(tmpdir):
    """Create sbws home inside of the test net tmp dir without initializing."""
    home = "/tmp/.sbws"
    os.makedirs(home, exist_ok=True)
    return home


@pytest.fixture(scope="session")
def sbwshome_dir(sbwshome_empty):
    """Create sbws home inside of the test net tmp dir without initializing."""
    os.makedirs(os.path.join(sbwshome_empty, "datadir"), exist_ok=True)
    return sbwshome_empty


@pytest.fixture(scope="session")
def test_config_path(tmpdir):
    config = tmpdir.join(".sbws.ini")
    return config


@pytest.fixture(scope="session")
def args(sbwshome_empty, parser, test_config_path):
    """Args with sbws home in the tests tmp dir."""
    args = _PseudoArguments(
        config=test_config_path,
        output=sbwshome_empty,
        scale=False,
        log_level="debug",
    )
    return args


@pytest.fixture(scope="session")
def conf(sbwshome_dir):
    """Default configuration with sbws home in the tmp test dir."""
    conf = _get_default_config()
    conf["paths"]["sbws_home"] = sbwshome_dir
    conf["paths"]["state_fpath"] = os.path.join(sbwshome_dir, "state.dat")
    conf["tor"]["run_dpath"] = os.path.join(sbwshome_dir, "tor", "run")
    conf["destinations"]["foo"] = "on"
    conf["destinations.foo"] = {}
    # The test server is not using TLS. Ideally it should also support TLS
    # If the url would start with https but the request is not using TLS,
    # the request would hang.
    conf["destinations.foo"]["url"] = "http://127.0.0.1:28888/sbws.bin"
    conf["tor"]["external_control_port"] = "8015"
    return conf


@pytest.fixture(scope="session")
def persistent_launch_tor(conf):
    cont = launch_or_connect_to_tor(conf)
    return cont


@pytest.fixture(scope="session")
def rl(args, conf, persistent_launch_tor):
    temp_rl = RelayList(args, conf, persistent_launch_tor)
    return temp_rl


@pytest.fixture(scope="session")
def cb(args, conf, persistent_launch_tor, rl):
    return CB(args, conf, persistent_launch_tor, rl)


@pytest.fixture(scope="session")
def dests(args, conf, persistent_launch_tor, cb, rl):
    dests, error_msg = DestinationList.from_config(
        conf, cb, rl, persistent_launch_tor
    )
    assert dests, error_msg
    return dests


@pytest.fixture(scope="session")
def is_cc_tor_version(persistent_launch_tor):
    version = persistent_launch_tor.get_version()
    print("version", version)
    return version >= Version("0.4.7.4-alpha-dev")



"""JSON custom serializers and deserializers."""
import datetime
import json

from .timestamps import DateTimeIntSeq, DateTimeSeq


class CustomEncoder(json.JSONEncoder):
    """JSONEncoder that serializes datetime to ISO 8601 string."""

    def default(self, obj):
        if isinstance(obj, DateTimeSeq) or isinstance(obj, DateTimeIntSeq):
            return [self.default(i) for i in obj.list()]
        if isinstance(obj, datetime.datetime):
            return obj.replace(microsecond=0).isoformat()
        else:
            return super().default(obj)


class CustomDecoder(json.JSONDecoder):
    """JSONDecoder that deserializes ISO 8601 string to datetime."""

    def decode(self, s, **kwargs):
        decoded = super().decode(s, **kwargs)
        return self.process(decoded)

    def process(self, obj):
        if isinstance(obj, list) and obj:
            return [self.process(item) for item in obj]
        if isinstance(obj, dict):
            return {key: self.process(value) for key, value in obj.items()}
        if isinstance(obj, str):
            try:
                return datetime.datetime.strptime(obj, "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                try:
                    datetime.datetime.strptime(
                        obj, "%Y-%m-%dT%H:%M:%S.%f"
                    ).replace(microsecond=0)
                except ValueError:
                    pass
            except TypeError:
                pass
        return obj



from unittest.mock import patch

from sbws.globals import RESULT_VERSION
from sbws.lib.relaylist import Relay
from sbws.lib.resultdump import (
    Result,
    ResultError,
    ResultErrorAuth,
    ResultErrorCircuit,
    ResultErrorStream,
    ResultSuccess,
    _ResultType,
)
from tests.unit.globals import monotonic_time


def test_Result(result):
    """
    A standard Result should not be convertible to a string because Result.type
    is not implemented.
    """
    try:
        str(result)
        print(str(result))
    except NotImplementedError:
        pass
    else:
        assert None, "Should have failed"


def test_Result_from_dict_bad_version():
    """
    The first thing that is checked is the version field, and a wrong one
    should return None
    """
    d = {"version": RESULT_VERSION + 1}
    r = Result.from_dict(d)
    assert r is None


def test_Result_from_dict_bad_type():
    """
    If the result type string doesn't match any of the known types, then it
    should throw NotImplementedError
    """
    d = {"version": RESULT_VERSION, "type": "NotARealType"}
    try:
        Result.from_dict(d)
    except NotImplementedError as e:
        assert str(e) == "Unknown result type NotARealType"
    else:
        assert None, "Should have failed"


def test_ResultSuccess_from_dict(result_success, result_success_dict):
    r2 = Result.from_dict(result_success_dict)
    assert isinstance(r2, ResultSuccess)
    assert str(result_success) == str(r2)


def test_ResultError_from_dict(result_error_stream, result_error_stream_dict):
    r2 = Result.from_dict(result_error_stream_dict)
    assert isinstance(r2, ResultError)
    assert str(result_error_stream) == str(r2)


@patch("time.time")
def test_ResultErrorCircuit(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorCircuit(relay, circ, dest_url, scanner_nick, msg=msg)
    r2 = ResultErrorCircuit(relay, circ, dest_url, scanner_nick, msg=msg, t=t)
    assert r1.msg == msg
    assert r1.nickname == nick
    assert r1.time == t
    assert r1.fingerprint == fp1
    assert r1.scanner == scanner_nick
    assert r1.type == _ResultType.ErrorCircuit
    assert r1.address == relay_ip
    assert r1.circ == circ
    assert r1.dest_url == dest_url
    assert r1.version == RESULT_VERSION
    assert str(r1) == str(r2)


@patch("time.time")
def test_ResultErrorCircuit_from_dict(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorCircuit(relay, circ, dest_url, scanner_nick, msg=msg)
    d = {
        "msg": msg,
        "fingerprint": fp1,
        "nickname": nick,
        "address": relay_ip,
        "circ": circ,
        "dest_url": dest_url,
        "scanner": scanner_nick,
        "version": RESULT_VERSION,
        "type": _ResultType.ErrorCircuit,
        "time": t,
        "master_key_ed25519": ed25519,
    }
    r2 = Result.from_dict(d)
    assert isinstance(r1, ResultErrorCircuit)
    assert isinstance(r2, ResultErrorCircuit)
    assert str(r1) == str(r2)


@patch("time.time")
def test_ResultErrorStream(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorStream(relay, circ, dest_url, scanner_nick, msg=msg)
    r2 = ResultErrorStream(relay, circ, dest_url, scanner_nick, msg=msg, t=t)
    assert r1.msg == msg
    assert r1.nickname == nick
    assert r1.time == t
    assert r1.fingerprint == fp1
    assert r1.scanner == scanner_nick
    assert r1.type == _ResultType.ErrorStream
    assert r1.address == relay_ip
    assert r1.circ == circ
    assert r1.dest_url == dest_url
    assert r1.version == RESULT_VERSION
    assert str(r1) == str(r2)


@patch("time.time")
def test_ResultErrorStream_from_dict(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorStream(relay, circ, dest_url, scanner_nick, msg=msg)
    d = {
        "msg": msg,
        "fingerprint": fp1,
        "nickname": nick,
        "address": relay_ip,
        "circ": circ,
        "dest_url": dest_url,
        "scanner": scanner_nick,
        "version": RESULT_VERSION,
        "type": _ResultType.ErrorStream,
        "time": t,
        "master_key_ed25519": ed25519,
    }
    r2 = Result.from_dict(d)
    assert isinstance(r1, ResultErrorStream)
    assert isinstance(r2, ResultErrorStream)
    assert str(r1) == str(r2)


@patch("time.time")
def test_ResultErrorAuth(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorAuth(relay, circ, dest_url, scanner_nick, msg=msg)
    r2 = ResultErrorAuth(relay, circ, dest_url, scanner_nick, msg=msg, t=t)
    assert r1.msg == msg
    assert r1.nickname == nick
    assert r1.time == t
    assert r1.fingerprint == fp1
    assert r1.scanner == scanner_nick
    assert r1.type == _ResultType.ErrorAuth
    assert r1.address == relay_ip
    assert r1.circ == circ
    assert r1.dest_url == dest_url
    assert r1.version == RESULT_VERSION
    assert str(r1) == str(r2)


@patch("time.time")
def test_ResultErrorAuth_from_dict(time_mock):
    t = 2000
    time_mock.side_effect = monotonic_time(start=t)
    fp1 = "A" * 40
    fp2 = "Z" * 40
    ed25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
    circ = [fp1, fp2]
    dest_url = "http://example.com/sbws.bin"
    scanner_nick = "sbwsscanner"
    nick = "Mooooooo"
    relay_ip = "169.254.100.1"
    relay = Result.Relay(fp1, nick, relay_ip, ed25519)
    msg = "aaaaayyyyyy bb"
    r1 = ResultErrorAuth(relay, circ, dest_url, scanner_nick, msg=msg)
    d = {
        "msg": msg,
        "fingerprint": fp1,
        "nickname": nick,
        "address": relay_ip,
        "circ": circ,
        "dest_url": dest_url,
        "scanner": scanner_nick,
        "version": RESULT_VERSION,
        "type": _ResultType.ErrorAuth,
        "time": t,
        "master_key_ed25519": ed25519,
    }
    r2 = Result.from_dict(d)
    assert isinstance(r1, ResultErrorAuth)
    assert isinstance(r2, ResultErrorAuth)
    assert str(r1) == str(r2)


def test_relay_in_recent_consensus_count(
    args, conf_results, controller, router_status, server_descriptor
):
    relay = Relay(
        router_status.fingerprint,
        controller,
        ns=router_status,
        desc=server_descriptor,
    )
    # Initialize the ResultSuccess as `measure_relay` does
    r = ResultSuccess(
        [],
        2000,
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    assert 1 == len(r.relay_in_recent_consensus)
    relay.update_relay_in_recent_consensus()
    r = ResultSuccess(
        [],
        2000,
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    assert 2 == len(r.relay_in_recent_consensus)


def test_relay_recent_measurement_attempt_count(
    args, conf_results, controller, router_status, server_descriptor
):
    relay = Relay(
        router_status.fingerprint,
        controller,
        ns=router_status,
        desc=server_descriptor,
    )
    relay.increment_relay_recent_measurement_attempt()
    relay.increment_relay_recent_measurement_attempt()
    # Initialize the ResultSuccess as `measure_relay` does
    r = ResultSuccess(
        [],
        2000,
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    assert 2 == len(r.relay_recent_measurement_attempt)


def test_relay_recent_priority_list_count(
    args, conf_results, controller, router_status, server_descriptor
):
    relay = Relay(
        router_status.fingerprint,
        controller,
        ns=router_status,
        desc=server_descriptor,
    )
    relay.increment_relay_recent_priority_list()
    relay.increment_relay_recent_priority_list()
    # Initialize the ResultSuccess as `measure_relay` does
    r = ResultSuccess(
        [],
        2000,
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    assert 2 == len(r.relay_recent_priority_list)
